/*
 *	BIRD -- Layer3 VPN Protocol Configuration
 *
 *	(c) 2011 Yandex, LLC
 *      Author: Alexander V. Chernikov <melifaro@yandex-team.ru>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

/**
 * DOC: l3vpn
 *
 * The l3vpn protocol works very much like pipe. 
 * It connects to VPNvX routing table (filled by RR bgp session) and local
 * 'client' one sending changes from one to other. update filters have to be
 * specified. Usually filtering is done via 'rt' attributes.
 *
 * To avoid loops, l3vpn keeps a `being updated' flag in each routing
 * table. l3vpn protocol is not intended to work with pipe protocol.
 */

#define LOCAL_DEBUG

#include "nest/bird.h"
#include "nest/iface.h"
#include "nest/protocol.h"
#include "nest/route.h"
#include "conf/conf.h"
#include "filter/filter.h"
#include "lib/string.h"
#include "lib/alloca.h"

#include "proto/bgp/bgp.h"
#include "l3vpn.h"

struct ldp_label_request
{
  int ltype;
  char *description;
};

struct ldp_label_release
{
  u32 label;
};

static u32
l3vpn_get_label(struct l3vpn_proto *p, net *n UNUSED)
{
  return p->label;
}

/**
 * l3vpn_write_labelstack - writes MPLS labelstack for L3vpn route
 * @me: pointer to destination labelstack data
 * @bgp_label: pointer to eattr with BA_MP_LABEL data
 * @nhop_label: pointer to eattr with LA_MPLSENTRY. Ca be NULL
 *
 */
static void
l3vpn_write_labelstack(struct bird_mpls_entry *me, eattr *bgp_label, eattr *nhop_label)
{
  struct bird_mpls_entry *nme = NULL;
  struct adata *ad;
  int i, j = 0;
  u32 *label;

  memset(me, 0, sizeof(struct bird_mpls_entry));

  /* Nexthop label can be NULL (LDP is not ready) */
  if (nhop_label != NULL)
  {
    ad = nhop_label->u.ptr;
    nme = (struct bird_mpls_entry *)(ad + 1);
    for (i = 0; i < MPLS_STACK_DEPTH; i++, j++)
    {
      me->stack[j] = nme->stack[i] & 0xFFFFF;
      if (nme->stack[i] & MPLS_STACK_BOS)
	break;
    }
  }

  /* BGP label can not be NULL */
  ad = bgp_label->u.ptr;
  for (label = (u32 *)(ad + 1); j < MPLS_STACK_DEPTH; label++, j++)
  {
    me->stack[j] = *label;
    if (*label & MPLS_STACK_BOS)
      break;
  }

  /* BOS already written by us, let's write the rest fields */
  me->action = KMPLS_OP_PUSH;
  me->depth = j + 1;
}

static void
l3vpn_rt_notify(struct proto *P, rtable *src_table, net *n, rte *new, rte *old, ea_list *attrs)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;
  rtable *dest;
  int order = (src_table == P->table) ? PE_TO_VPN : VPN_TO_PE;
  struct announce_hook *h;
  struct proto_stats *stats, *old_stats;
  struct proto *src;
  ea_list *tmpa, *eal;
  eattr *ea;
  eattr *bgp_label, *nhop_label;
  struct adata *ad;
  struct hostentry *he;
  char buf[32];

  vpn4_addr vpn;
  void *addr;

  net *nn;
  rte *e;
  rta *newa;

  if (!new && !old)
    return;

  /* Do address conversion */
  if (order == VPN_TO_PE)
  {
    memcpy(&vpn, FPREFIX(&n->n), src_table->fib.addr_size);
    /* Search for IP address */
    addr = &vpn.addr;
    dest = P->table;
    h = P->thook;
    stats = &p->vpn_stats;
  }
  else
  {
    /* PE -> VPN */
    vpn.rd = p->rd;
    memcpy(&vpn.addr, FPREFIX(&n->n), src_table->fib.addr_size);
    /* Search for VPN address */
    addr = &vpn;
    dest = p->vpn_table;
    h = p->a;
    stats = &P->stats;
  }

  if (dest->pipe_busy)
  {
    log(L_ERR "l3vpn loop detected when sending %F to table %s", &n->n, dest->name);
    return;
  }

  addr_print(buf, sizeof(buf), dest->fib.addr_type, addr);
  log(L_ERR "Converting %F -> %s/%d", &n->n, buf, n->n.pxlen);

  nn = fib_get(&dest->fib, addr, n->n.pxlen);

  if (new)
  {
    /* 
     * Regadless of direction we have to add 1 attribute:
     * for PE_TO_VPN we're adding BA_MP_LABEL
     * for VPN_TO_PE we're adding LA_MPLSENTRY
     */
    tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
    memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
    tmpa->count = 1;

    ea = &tmpa->attrs[0];

    /* Setup external storage for attribute */
    ad = alloca(sizeof(struct adata) + MAX(sizeof(struct bird_mpls_entry), sizeof(u32)));
    ea->u.ptr = ad;

   /* 
    * Client route, we should pass all base/extended attributes (which are not filtered out)
    * Filtering attributes (like BGP communities passed from client) is a filter task
    * Add MPLS label if needed 
    */
    if (order == PE_TO_VPN)
    {
      /* Prepare BA_MP_LABEL attribute data */
      ad->length = sizeof(u32);
      /* Some BGP internals have to be used here */
      ea->id = EA_CODE(EAP_BGP, BA_MP_LABEL);
      ea->type = EAF_TYPE_INT_SET;
      ea->flags = BAF_OPTIONAL;

      /* Let's write label */
      u32 *label = (u32 *)ad->data;
      *label = l3vpn_get_label(p, n) | MPLS_STACK_BOS;
    }
   else /* VPN_TO_PE */
    {
      ad->length = sizeof(struct bird_mpls_entry);
      memset(ad->data, 0, sizeof(struct bird_mpls_entry));
      /* Some LDP internals have to be used here */
      ea->id = EA_CODE(EAP_LDP, LA_MPLSENTRY);
      ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;

      /* 
       * Nexthop resolution is already done in BGP code (this is mandatory to 
       * make best path algorithm to work). We got valid nexthop address in
       * new->attrs and original route attributes in new->attrs->hostentry->source
       *
       * Nexthop update is handled in BGP code, too. We get route update event,
       * so there is no need to make any recursive resolving here
       */

      /*
       * Find mandatory BA_MP_LABEL attribute in VPNvX route attribute
       */
      if ((bgp_label = ea_find(attrs, EA_CODE(EAP_BGP, BA_MP_LABEL))) == NULL)
      {
        DBG("No BA_MP_LABEL in converted VPNvX route (%F), skipping\n", &n->n);
        return;
      }

      if ((he = new->attrs->hostentry) == NULL)
      {
	DBG("Nexthop for route %F is not recursive\n", &n->n);
	nhop_label = NULL;
      }
     else
      {
	if ((nhop_label = ea_find(he->src->eattrs, EA_CODE(EAP_LDP, LA_MPLSENTRY))) == NULL)
	{
	  /* 
	   * This can be perfectly valid situation - if this nexthop is directly connected to
	   * us and sends Ipmplicit null for its loopback address.
	   */
	  DBG(" No labelstack attribute for route to %F via %I, this can possibly cause blackhole\n", &n->n, he->addr);
	}
      }
    
      l3vpn_write_labelstack((struct bird_mpls_entry *)ad->data, bgp_label, nhop_label);
    }

    /* Make new attributes */
    eal = new->attrs->eattrs;
    /* Copy with (possibly) modified eattrs */
    new->attrs->eattrs = attrs;
    newa = rta_copy_with_ea(new->attrs, tmpa);
    /* Put original pointer back  */
    new->attrs->eattrs = eal;
    newa->aflags = 0;
    newa->hostentry = NULL;

    e = rte_get_temp(newa);
    e->net = nn;
    e->pflags = 0;

    /* Copy protocol specific embedded attributes. */
    memcpy(&(e->u), &(new->u), sizeof(e->u));
    e->pref = new->pref;
    e->pflags = new->pflags;

    src = new->attrs->proto;
   /* Finally, do attribute lookup */
   rta_lookup_self(newa);
  }
  else
  {
    e = NULL;
    src = old->attrs->proto;
  }

  src_table->pipe_busy = 1;
  /*
   * FIXME: Trick with stats pointer is needed to produce
   * valid numbers on protocol shutdown.
   */
  old_stats = h->stats;
  h->stats = stats;
  do_rte_update(dest, h, nn, &p->p, src, e);
  h->stats = old_stats;
  src_table->pipe_busy = 0;
}

static int
l3vpn_import_control(struct proto *P, rte **ee, ea_list **ea UNUSED, struct linpool *p UNUSED)
{
  struct proto *pp = (*ee)->sender;

  if (pp == P)
    return -1;	/* Avoid local loops automatically */

  return 0;
}

static int
l3vpn_reload_routes(struct proto *P)
{
  /*
   * Because the l3vpn protocol feeds routes from both routing tables
   * together, both directions are reloaded during refeed and 'reload
   * out' command works like 'reload' command. For symmetry, we also
   * request refeed when 'reload in' command is used.
   */
  proto_request_feeding(P);
  return 1;
}

/**
 * l3vpn_run - starts instance, connect to routing tables
 * @p: pointer to protocol instance
 *
 * protocol state is not changed, proto_notify_state()
 * needs to be called
 */
static void
l3vpn_run(struct l3vpn_proto *p)
{
  DBG("l3vpn_run() for %s\n", p->p.name);
  struct announce_hook *a;

  /* Clean up the secondary stats */
  bzero(&p->vpn_stats, sizeof(struct proto_stats));

  /* Lock the peer table, unlock is handled in l3vpn_shutdown() */
  rt_lock_table(p->vpn_table);
  rt_lock_table(p->mpls_table);

  /* Connect the protocol to vpn table. */
  a = proto_add_announce_hook(&p->p, p->vpn_table, RA_OPTIMAL);
  a->stats = &p->vpn_stats;
  p->a = a;
}

/**
 * l3vpn_rel_notify - dispatches forced label release
 * @protived: protocol withdrawing service
 * @requester: our protocol
 * @repl: label information
 */
void
l3vpn_rel_notify(struct proto *provider, struct proto *requester, struct service_reply *repl)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *)requester;

  DBG("l3vpn_rel_notify() called from %s to %s t/st %d/%d table %s\n", 
	provider->name, requester->name, repl->request, repl->subclass, repl->t ? repl->t->name : "NULL");

  if ((repl->request != SRCLASS_LABEL) || (repl->subclass != 1))
    return;
/*
  if (repl->t != p->mpls_table) 
    return;
*/
  if (*((u32 *)(repl + 1)) != p->label)
    return;

  p->label = 0;

  /* Valid LDP protocol requests us to free label and cause shutdown */
  proto_notify_state(&p->p, PS_DOWN);
}

static int
l3vpn_request_vpn_label(struct l3vpn_proto *p)
{
  struct ldp_label_request *lreq;
  char dbuf[64];
  byte rbuf[sizeof(struct service_reply) + sizeof(u32)], *pbuf;
  size_t buflen;
  u32 *label;

  struct service_request *req = alloca(sizeof(struct service_request) + sizeof(struct ldp_label_request));

  /* Prepare description */
  bsnprintf(dbuf, sizeof(dbuf), "VRF %s", p->p.name);

  lreq = (struct ldp_label_request *)(req + 1);
  lreq->ltype = 1; /* LDP_LIB_TYPE_L3VPN */
  lreq->description = dbuf;

  req->request = SRCLASS_LABEL;
  req->subclass = 1; /* SRSUB_LABEL_REQ */
  req->t = p->mpls_table;
  req->p = &p->p;
  req->target = NULL;
  req->relfunc = l3vpn_rel_notify;

  buflen = sizeof(rbuf);
  pbuf = rbuf;

  DBG("Requesting label..\n");

  if (proto_request_service(req, &pbuf, &buflen) != SR_OK)
    return 0;

  label = (u32 *)(pbuf + sizeof(struct service_reply));

  DBG("Got label %u for L3VPN %s\n", *label, p->p.name);

  p->label = *label;

  return 1;
}

static void
l3vpn_free_vpn_label(struct l3vpn_proto *p, u32 label)
{
  struct ldp_label_release *lrel;
  byte *pbuf;
  size_t buflen;
  struct service_request *req = alloca(sizeof(struct service_request) + sizeof(struct ldp_label_release));

  lrel = (struct ldp_label_release *)(req + 1);
  lrel->label = label;

  req->request = SRCLASS_LABEL;
  req->subclass = 2; /* SRSUB_LABEL_REL */
  req->t = p->mpls_table;
  req->p = &p->p;
  req->target = NULL;

  buflen = 0;
  pbuf = NULL;

  DBG("Releasing VPN label.. %u\n", label);

  proto_request_service(req, &pbuf, &buflen);
}

static void
ld3vpn_check_ldp(timer *t)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *)t->data;

  if (l3vpn_request_vpn_label(p))
  {
    /* Found LDP protocol instance, let's start up */
    l3vpn_run(p);
    proto_notify_state(&p->p, PS_UP);
    return;
  }

  /* Not found. restart timer */
  tm_start(t, 10);
}

static int
l3vpn_start(struct proto *P)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;
  timer *t;

  DBG("l3vpn_start() called for %s\n", P->name);

  t = p->request_timer = tm_new(P->pool);
  t->hook = ld3vpn_check_ldp;
  t->data = p;

  /* Clean up the secondary table stats */
  bzero(&p->vpn_stats, sizeof(struct proto_stats));

  /* Request label */
  if (l3vpn_request_vpn_label(p) == 0)
  {
    /* 
     * We should try to re-request label from LDP periodically.
     * .. and show custom protocol status in CLI (like BGP do)
     */
    t->randomize = 4;
    /* Start periodic LDP protocol checks */
    tm_start(t, 10);
    return PS_START;
  }

  /* run protocol */
  l3vpn_run(p);

  return PS_UP;
}

static int
l3vpn_shutdown(struct proto *P)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;

  DBG("l3vpn_shutdown() called for %s\n", P->name);

  if (P->proto_state == PS_UP)
  {
    DBG("Freeing tables..\n");
    rt_unlock_table(p->mpls_table);
    rt_unlock_table(p->vpn_table);
  }

  rfree(p->request_timer);
  p->request_timer = NULL;

  if (p->label > 0)
  {
    l3vpn_free_vpn_label(p, p->label);
    p->label = 0;
  }

  return PS_DOWN;
}

static struct proto *
l3vpn_init(struct proto_config *C)
{
  struct l3vpn_config *c = (struct l3vpn_config *) C;
  struct proto *P = proto_new(C, sizeof(struct l3vpn_proto));
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;

  p->vpn_table = c->vpn_table->table;
  p->mpls_table = c->mpls_table->table;
  p->rd = c->rd;
  P->accept_ra_types = RA_OPTIMAL;
  P->rt_notify = l3vpn_rt_notify;
  P->import_control = l3vpn_import_control;
  P->reload_routes = l3vpn_reload_routes;

  return P;
}

static void
l3vpn_postconfig(struct proto_config *C)
{
  struct l3vpn_config *c = (struct l3vpn_config *) C;

  if (!c->vpn_table)
    cf_error("Name of vpn table not specified");

  if (!c->mpls_table)
    cf_error("Name of mpls table not specified");

  if ((c->vpn_table->addr_type != RT_VPNV4) && (c->vpn_table->addr_type != RT_VPNV6))
    cf_error("VPN Table %s type is not VPN", c->vpn_table->name);

  if (c->mpls_table->addr_type != RT_MPLS)
    cf_error("MPLS Table %s type is not MPLS", c->mpls_table->name);

  if ((C->table->addr_type != RT_IPV4) && (C->table->addr_type != RT_IPV6))
    cf_error("Client table %s type is not IPv4/IPv6", C->table->name);

  if (((c->vpn_table->addr_type == RT_VPNV4) && (C->table->addr_type != RT_IPV4)) ||
	  ((c->vpn_table->addr_type == RT_VPNV6) && (C->table->addr_type != RT_IPV6)))
    cf_error("Address families are different for VPN table %s and client table %s", c->vpn_table->name, C->table->name);

  /* 
   * No checks for IGP table is done since VPNv6 table can containg IPv4 nexthops
   */
}

static void
l3vpn_get_status(struct proto *P, byte *buf)
{
  char rdbuf[20];
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;

  switch (P->proto_state)
  {
    case PS_START:
      bsprintf(buf, "LDP query..");
      break;
    case PS_UP:
      addr_print_rd(rdbuf, sizeof(rdbuf), p->rd);
      bsprintf(buf, "%s <[%s]> %s", P->table->name, rdbuf, p->vpn_table->name);
      break;
    default:
      buf[0] = '\0';
  }
}

static int
l3vpn_reconfigure(struct proto *P, struct proto_config *new)
{
  struct announce_hook *a, *pa;
  struct l3vpn_proto *p = (struct l3vpn_proto *)P;
  struct l3vpn_config *o = (struct l3vpn_config *) P->cf;
  struct l3vpn_config *n = (struct l3vpn_config *) new;

  if (((o->vpn_table->table != n->vpn_table->table)) || (o->rd != n->rd))
    return 0;

  if (o->mpls_table->table != n->mpls_table->table)
    return 0;

  /*
   * reconfigure can be called in any protocol status
   */
  if (!((a = P->thook) && (pa = p->a)))
    return 1;

  /* Update filters */
  /* See pipe.c comments for full explanation */

  pa->out_filter = a->in_filter;
  a->in_filter = FILTER_ACCEPT;
  pa->in_filter = FILTER_ACCEPT;
  return 1;
}


struct protocol proto_l3vpn = {
  name:		"l3vpn",
  template:	"l3vpn%d",
  postconfig:	l3vpn_postconfig,
  init:		l3vpn_init,
  start:	l3vpn_start,
  shutdown:	l3vpn_shutdown,
  reconfigure:	l3vpn_reconfigure,
  get_status:	l3vpn_get_status,
};
