/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * LDP cli functions and various related helpers
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"
#include "nest/cli.h"

void
ldp_get_route_info(rte *e, byte *buf, ea_list *attrs)
{
	eattr *ea;
	struct adata *ad;
	bird_mpls_entry *me;
	char lbuf[32];

	if ((ea = ea_find(attrs, EA_CODE(EAP_LDP, LA_MPLSENTRY))) == NULL)
		return;

	ad = ea->u.ptr;
	me = (bird_mpls_entry *)(ad + 1);

	addr_print_me(lbuf, sizeof(lbuf), me, 0);

  	buf += bsprintf(buf, " %s (%s)", ldp_lib_texttype(e->u.ldp.ltype), lbuf);

	if (e->u.ldp.router_id)
		buf += bsprintf(buf, " [%R]", e->u.ldp.router_id);
}

/*
 * Shows protocol extended attributes in human-readable form.
 * Base protocol hook
 */
int
ldp_get_attr(eattr *a, byte *buf, int buflen)
{
	unsigned int i = EA_ID(a->id);
	int len;
	struct adata *ad;
	struct bird_mpls_entry *me;

	switch (i)
	{
		case LA_MPLSENTRY:
			ad = (struct adata *)a->u.ptr;
			me = (struct bird_mpls_entry *)ad->data;
			len = bsprintf(buf, "%s: [%s] %s", "label", ldp_label_texttype(me->local_label), ldp_mplsop_texttype(me->action));
			buf += len;
			buflen -= len;
			if (me->depth > 0)
			{
				len = bsprintf(buf, " { ");
				for (i = 0; i < me->depth; i++)
					len += bsprintf(buf + len, "%s, ", ldp_label_texttype(me->stack[i] & 0xFFFFF));
				len -= 2;
				len += bsprintf(buf + len, " }");
			}

			return GA_FULL;
	}

	return GA_UNKNOWN;
}

#define CLI_OUT(fmt, ...)	cli_msg(-1006, fmt, ##__VA_ARGS__)
#define CLI_END()		cli_msg(0, "")
void
ldp_show_neighbour(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_iface *li;
	ldp_hello *lh, *lh_next;
	ldp_neighbour *ln;
	ldp_session *ls, *ls_next;
	time_t tm;

	if (pl == NULL)
	{
		CLI_OUT("Please specify protocol instance");
		return;
	}

	if (strcmp(p->proto->name, proto_ldp.name))
	{
		CLI_OUT("Incorrect protocol type: got %s instead of %s", p->proto->name, proto_ldp.name);
		return;
	}

/*
    Peer LDP Ident: 10.0.0.88:0; Local LDP Ident 10.2.33.4:0
        TCP connection: 10.0.0.88.646 - 10.2.33.4.11018
        State: Oper; Msgs sent/rcvd: 18/9; Downstream
        Up time: 00:00:07
        LDP discovery sources:
          FastEthernet0/0.5, Src IP addr: 10.1.5.88
*/
	WALK_LIST_DELSAFE(ls, ls_next, pl->session_list)
	{
		ls = SKIP_BACK(ldp_session, n_list, ls);
		ln = ls->ln;

		CLI_OUT("    Peer LDP Ident: %I:%d; Local LDP Ident %I:%d", ln->lsr_id, ls->space_id, pl->router_id, 0);
		if ((ln->sk != NULL) && (ls->state >= LDP_SSM_INITIALIZED))
		{
			CLI_OUT("         TCP connection: %I.%d - %I.%d", ln->sk->daddr, ln->sk->dport, ln->sk->faddr, 0);
			CLI_OUT("         State: %s; Msgs sent/rcvd: %d/%d; %s",
					ldp_session_textstatus(ls->state), ls->messages_out, ls->messages_in, "Downstream");
			tm = CurrentTime - ls->last_state;
			CLI_OUT("         Up time: %02d:%02d:%02d", tm / 3600, (tm % 3600) / 60, tm % 60);
		}

		CLI_OUT("         LDP discovery sources:");
		WALK_LIST_DELSAFE(lh, lh_next, ln->hello_list)
		{
			lh = SKIP_BACK(ldp_hello, n_lsr, lh);
			li = lh->li;
			CLI_OUT("           %s, Src IP addr: %I", li->iface->name, lh->remote);
		}

	}
	CLI_END();
}

void
ldp_show_interfaces(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_iface *li;
	struct iface *i;
	//ldp_hello *lh;
	//ldp_neighbour *ln;
	//time_t tm;
	char buf[50];

	if (pl == NULL)
	{
		CLI_OUT("Please specify protocol instance");
		goto done;
	}

	if (strcmp(p->proto->name, proto_ldp.name))
	{
		CLI_OUT("Incorrect protocol type: got %s instead of %s", p->proto->name, proto_ldp.name);
		goto done;
	}

/*
    Interface em0, IP addr: 10.1.5.88
        State: Operational
	LDP neighbours:
	  Ident: 10.0.0.88:0; From: 10.2.33.4
*/
	WALK_LIST(li, pl->iface_list)
	{
		i = li->iface;

		if (i->addr != NULL)
			bsnprintf(buf, sizeof(buf), "%I/%d", i->addr->ip, i->addr->pxlen);
		else
			bsnprintf(buf, sizeof(buf), "%s", "<UNAVAIL>");
		CLI_OUT("    Interface %s, IP addr: %s", i->name, buf);
		CLI_OUT("         State: %s; bird state: %s", ldp_iface_textstatus(li->status), (i->flags & IF_UP) ? "UP" : "DOWN");
		CLI_OUT("         LDP discovery sources:");
#if 0
		WALK_LIST(lh, pl->hello_list)
		{
			lh = SKIP_BACK(ldp_hello, n_list, lh);
			if (lh->li != li)
				continue;
			CLI_OUT("           %s, Src IP addr: %I", i->name, lh->remote);
		}
#endif
	}
done:
	CLI_END();
}

void
ldp_show_bindings(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	ldp_session *ls;
	ldp_rib_group *rg;
	ldp_rib_entry *re;

	if (pl == NULL)
	{
		CLI_OUT("Please specify protocol instance");
		return;
	}

	if (strcmp(p->proto->name, proto_ldp.name))
	{
		CLI_OUT("Incorrect protocol type: got %s instead of %s", p->proto->name, proto_ldp.name);
	}
/*
  lib entry: 0.0.0.0/0, rev 14
        local binding:  label: imp-null
  lib entry: 10.0.0.0/24, rev 56
        local binding:  label: 20
        remote binding: lsr: 10.0.0.88:0, label: exp-null
  lib entry: 10.0.0.88/32, rev 16
        local binding:  label: 17
        remote binding: lsr: 10.0.0.88:0, label: exp-null
  lib entry: 10.1.4.0/24, rev 10
        local binding:  label: imp-null
  lib entry: 10.1.5.0/24, rev 12
        local binding:  label: imp-null
        remote binding: lsr: 10.0.0.88:0, label: exp-null
*/
	FIB_WALK(&pl->rib, f)
	{
		rg = (ldp_rib_group *)f;
		if ((re = (ldp_rib_entry *)HEAD(rg->entries)) == NULL)
			continue;

		if ((re->flags & LDP_RE_BEST) == 0)
			continue;
/*
		if (re->ls != NULL)
			continue;
*/
		CLI_OUT("  lib entry: %I/%d", rg->n.prefix, rg->n.pxlen);
		if (re->flags & LDP_RE_HASLABEL)
			CLI_OUT("      local binding:  label: %s", ldp_label_texttype(re->label));

		WALK_LIST(re, rg->entries)
		{
			if ((ls = re->ls) != NULL)
				CLI_OUT("      remote binding:  lsr: %I:%d, label: %s", 
						ls->ln->lsr_id, ls->space_id, ldp_label_texttype(re->me.stack[0]));
		}
	}
	FIB_WALK_END;

	CLI_END();
}

void
ldp_show_forwardingtable(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	ldp_lib_general *lg;
	ldp_lib_transit *lt;
	char abuf[50], *x;

	if (pl == NULL)
	{
		CLI_OUT("Please specify protocol instance");
		return;
	}

	if (strcmp(p->proto->name, proto_ldp.name))
	{
		CLI_OUT("Incorrect protocol type: got %s instead of %s", p->proto->name, proto_ldp.name);
	}
/*
Local  Outgoing      Prefix            Bytes Label   Outgoing   Next Hop
Label  Label or VC   or Tunnel Id      Switched      interface
16     No Label      10.2.0.2/32       0             As1        point2point
17     Pop Label     10.0.0.88/32      0             Fa0/0.5    10.1.5.88
18     Pop Label     10.1.5.88/32      0             Fa0/0.5    10.1.5.88
19     No Label      10.1.6.0/24       0             Fa0/0.5    10.1.5.88
20     No Label      10.0.0.0/24       0             Fa0/0.5    10.1.5.88
21     No Label      10.2.0.0/24       0             Fa0/0.5    10.1.5.88
22     No Label      10.2.33.3/32      0             Fa0/0.5    10.1.5.88
23     No Label      10.1.6.88/32      0             Fa0/0.5    10.1.5.88
*/
	CLI_OUT("%-6s %-14s %-18s %-14s %-10s %-8s", "Local", "Outgoing", "Prefix", "Bytes Label", "Outgoing", "Next Hop");
	CLI_OUT("%-6s %-14s %-18s %-14s %-10s", "Label", "Label or VC", "or Tunnel Id", "Switched", "interface");

	FIB_WALK(&pl->lib->f, ltmp)
	{
		lg = (ldp_lib_general *)ltmp;

		switch(lg->type)
		{
			case LDP_LIB_TYPE_TRANSIT:
				lt = (ldp_lib_transit *)lg;
				ip_ntop(lt->prefix, abuf);
				x = abuf + strlen(abuf);
				sprintf(x, "/%d", lt->pxlen);

				CLI_OUT("%-6d %-14s %-18s %-14d %-10s %-8I", lg->me.local_label, 
					ldp_mplsop_texttype(lg->me.action), abuf, 0, "?", lt->nexthop);
				break;

			default:
				CLI_OUT("%-6d %-14s %-18s %-14d %-10s %-8I", lg->me.local_label, 
					ldp_mplsop_texttype(lg->me.action), "unkn", 0, "?", "?");
		}
	}
	FIB_WALK_END;

	CLI_END();
}

