/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * Various databases are mananaged hered:
 * LFIB (Local label mappings)
 * RIB (Remote label bindings)
 * Addrmap (Nexthop <> neighbour mapping)
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"

inline int ldp_addrmap_hash(ip_addr addr);
static u32 ldp_lib_hash(struct fib *f, void *a);

static void
ldp_rib_init_prefix(struct fib_node *n)
{
	ldp_rib_group *rg = (ldp_rib_group *)n;

	init_list(&rg->entries);
	rg->flags = LDP_RG_NEW;
}

int
ldp_rib_init(struct proto_ldp *pl)
{
	fib_init(&pl->rib, MAIN_POOL(pl), sizeof(struct _ldp_rib_group), 0, ldp_rib_init_prefix);

	return LDP_OK;
}

void
ldp_rib_free(struct proto_ldp *pl, int last UNUSED)
{
	ldp_rib_group *rg;
	ldp_rib_entry *re;

	/* Free all additionally allocated stuff */
	FIB_WALK(&pl->rib, rtmp)
	{
		rg = (ldp_rib_group *)rtmp;

		WALK_LIST_FIRST(re, rg->entries)
		{
			rem_node(NODE re);
			mb_free(re);
		}
	}
	FIB_WALK_END;

	fib_free(&pl->rib);

	return;
}

/**
 * ldp_rib_add_entry - Adds FEC prefix to given database
 * @pl: pointer to protocol instance
 * @prefix: prefix IP address
 * @pxlen: prefix mask
 * @ls: pointer to ldp session (if exists)
 *
 * Adds RIB group entry if not exists
 * Adds RIB entry to group entry list and neighbour prefix list
 *
 * Returns pointer to created RIB entry or NULL in case of error
 */
ldp_rib_entry *
ldp_rib_add_entry(struct proto_ldp *pl, ip_addr prefix, int pxlen, ldp_session *ls)
{
	ldp_rib_group *rg;
	ldp_rib_entry *re;

	if ((re = mb_allocz(MAIN_POOL(pl), sizeof(ldp_rib_entry))) == NULL)
		return NULL;

	rg = fib_get(&pl->rib, &prefix, pxlen);

	if (rg->flags & LDP_RG_NEW)
	{
		/* New node, reinit */
		rg->flags &= ~LDP_RG_NEW;
	}
	
	/* Prepare new entry to be added */
	re->ls = ls;
	re->rg = rg;

	/* Add to per-session list */
	if (ls != NULL)
	{
		LOG_DB_DBG("++ RIB %I/%d in session %R:%d", prefix, pxlen, ls->ln->lsr_id, ls->space_id);
		add_tail(&ls->prefix_list, &re->n_lss);
	}
	else
		LOG_DB_DBG("++ RIB %I/%d", prefix, pxlen);

	/* Add to prefix group list */
	add_tail(&rg->entries, NODE re);
	
	return re;
}

/**
 * ldp_rib_set_best - mark RIB entry as best (e.g. installed)
 * @pl: pointer to protocol instance
 * @lrd: pointer to route distinguisher
 * @re: pointer to ldp session if exists
 *
 * Moves RIB entry on the first place in appropriate group 
 * and marks it as BEST
 *
 */
void
ldp_rib_set_best(struct proto_ldp *pl, ldp_rib_entry *re)
{
	re->flags |= LDP_RE_BEST;
	rem_node(NODE re);
	add_head(&re->rg->entries, NODE re);
}

/**
 * ldp_rib_del_entry - Deletes FEC prefix from RIB database
 * @pl: pointer to protocol instance
 * @re: pointer to RIB entry
 *
 * Deletes RIB entry from group entry list and from neighbour prefix list
 * Drops RIB group entry if no entries remains
 *
 * Returns LDP_OK in case of successful deletion or LDP_ERR overwise
 */
int
ldp_rib_del_entry(struct proto_ldp *pl, ldp_rib_entry *re)
{
	ldp_rib_group *rg = re->rg;
	ldp_session *ls = re->ls;

	/* Remove from session prefix list */
	if (ls != NULL)
		rem_node(&re->n_lss);

	rem_node(NODE re);
	mb_free(re);

	if (ls != NULL)
		LOG_DB_DBG("-- RIB %I/%d in session %I:%d", rg->n.prefix, rg->n.pxlen, ls->ln->lsr_id, ls->space_id);
	else
		LOG_DB_DBG("-- RIB %I/%d", rg->n.prefix, rg->n.pxlen);

	/* Check if we need to free rib group */
	if (EMPTY_LIST(rg->entries))
		fib_delete(&pl->rib, rg);

	return LDP_OK;
}

ldp_rib_group *
ldp_rib_search_group(struct proto_ldp *pl, ip_addr prefix, int pxlen)
{
	return fib_find(&pl->rib, &prefix, pxlen);
}

ldp_rib_entry *
ldp_rib_search_entry(struct proto_ldp *pl, ip_addr prefix, int pxlen, ldp_session *ls)
{
	ldp_rib_group *rg;
	ldp_rib_entry *re;

	if ((rg = ldp_rib_search_group(pl, prefix, pxlen)) == NULL)
		return NULL;

	WALK_LIST(re, rg->entries)
	{
		if (re->ls == ls)
			return re;
	}

	return NULL;
}

static void
ldp_lib_init_prefix(struct fib_node *n)
{
	ldp_lib_general *lg = (ldp_lib_general *)n;
	memset(&lg->me, 0, LDP_LIB_ENTRY_SIZE - OFFSETOF(ldp_lib_general, me));
}
/*
 * Initialize LIB database
 */
int
ldp_lib_init(struct proto_ldp *pl)
{
	ldp_lib_database *lib;

	if ((lib = mb_alloc(MAIN_POOL(pl), sizeof(ldp_lib_database))) == NULL)
	{
		LOG_ERR("LIB database initialization failed!");
		return LDP_ERR;
	}

	fib2_init(&lib->f, MAIN_POOL(pl), LDP_LIB_ENTRY_SIZE, RT_MPLS, sizeof(u32), 0, ldp_lib_init_prefix, ldp_lib_hash);

	lib->current_label = pl->label_min;
	init_list(&lib->free_items);

	pl->lib = lib;

	return LDP_OK;
}

static u32
ldp_lib_hash(struct fib *f, void *a)
{
	return (*((u32 *)a) & 0xFFFFF);
}

/*
 * Free LIB database.
 * XXX: we have to notify 'foreign' label requestors about
 * shutdown taken place.
 */
void
ldp_lib_free(struct proto_ldp *pl, int last UNUSED)
{
	ldp_lib_database *lib = pl->lib;
	u32 label;
	ldp_lib_general *lg;
	ldp_lib_free_label *fl, *fl_next;
	struct service_reply *repl = alloca(sizeof(struct service_reply) + sizeof(u32));
	u32 *plabel;

	/* Prepare common fields for repl */
	repl->request = SRCLASS_LABEL;
	repl->subclass = SRSUB_LABEL_REQ;
	repl->t = pl->mpls_table;
	repl->p = &pl->proto;
	plabel = (u32 *)repl->data;

	FIB_WALK(&lib->f, ltmp)
	{
		lg = (ldp_lib_general *)ltmp;
		label = *((u32 *)FPREFIX(&lg->n));
		switch (lg->type)
		{
			case LDP_LIB_TYPE_TRANSIT:
				// local label
				break;
			default:
				/* Notify requester we're shutting down */
				if (lg->relfunc)
				{
					*plabel = lg->me.local_label;
					lg->relfunc(&pl->proto, lg->proto, repl);
				}
					
				break;
		}

		/* Do nothing for now since actual freeing will be done in fib_free() */
		//mb_free(lg);
	}
	FIB_WALK_END;

	WALK_LIST_DELSAFE(fl, fl_next, lib->free_items)
	{
		rem_node(NODE fl);
		mb_free(fl);
	}

	/* Free database fib/itself */
	fib_free(&lib->f);
	mb_free(lib);
	pl->lib = NULL;
}
/**
 * ldp_lib_label_add - Allocate new LDP label of given type from LDP database
 * @pl: pointer to LDP protocol
 * @type: LIB entry type
 * @op: MPLS operation to take
 *
 * Caller has to fill most fields by itself, database care of general
 * label allocation/deallocation only.
 * If label is requested by foreign caller, ->info field has to be filled
 * by human-readable allocation data
 *
 * Returns: ldp_lib_(general|transit) structure or NULL in case of failure
 */
ldp_lib_general *
ldp_lib_label_add(struct proto_ldp *pl, enum LDP_LIB_TYPE type, enum kmpls_op op)
{
	u32 label = 0;
	ldp_lib_general *lg;
	ldp_lib_database *lib = pl->lib;
	ldp_lib_free_label *fl, *fl_next;

	/* 
	 * Find free label number to allocate
	 * Labels are allocated in ascending order.
	 * We keep track of maximum allocated number and
	 * all free labels in pl->label_min .. lib->current_label - 1 space
	 *
	 * On label allocation request we try to eliminate all free items at first
	 * (taking care on possible changed config label space)
	 * IFF free items list is empty we try to increment current_label counter
	 *
	 * On label release we check if we can decrement current_label counter 
	 * (in case of last allocated label is released) and in case of failure - 
	 * push free label in free label list
	 */

	/* Check if configured label space has changed */
	if (lib->current_label < pl->label_min)
		lib->current_label = pl->label_min;

	if (!EMPTY_LIST(lib->free_items))
	{
		/* Walk through free labels list */
		WALK_LIST_DELSAFE(fl, fl_next, lib->free_items)
		{
			if ((fl->label < pl->label_min) || (fl->label > pl->label_max))
			{
				/* Not available in current labelspace */
				rem_node(NODE fl);
				mb_free(fl);
				continue;
			}

			/* Found */
			label = fl->label;
			rem_node(NODE fl);
			mb_free(fl);
			LOG_DB_DBG("Found label %d in empty space", label);
			break;
		}
	}

	/* If label is still not found => free items list is (now) empty */
	if (label == 0)
	{
		if (pl->lib->current_label > pl->label_max)
		{
			LOG_DB_ERR("No free labels left, consider increasing label_max");
			return NULL;
		}

		label = pl->lib->current_label++;
	}

	LOG_DB_DBG("Allocated type/%s label %d", ldp_lib_texttype(type), label);

	/* Request label from FIB */
	lg = fib_get(&lib->f, &label, 0);

	/* Fill in all data we know */
	lg->type = type;
	lg->me.action = op;
	lg->me.depth = 1;
	lg->me.local_label = label;
	
	return lg;
}

/**
 * ldp_lib_label_del - Deletes label from label database
 * @pl: pointer to protocol 
 * @label: MPLS label number
 *
 * Finds and deletes label from LDP database
 * SIDE EFFECTS:
 * Depending on label types, label structure is removed from appropriate lists, too
 *
 */
int
ldp_lib_label_del(struct proto_ldp *pl, u32 label)
{
	ldp_lib_general *lg;
	ldp_lib_free_label *fl;
	ldp_lib_transit *lt;
	ldp_lib_database *lib = pl->lib;

	LOG_DB_ERR("Freeing label %u", label);

	if ((lg = fib_find(&lib->f, &label, 0)) == NULL)
	{
		LOG_DB_ERR("Label %u not found in LIB", label);
		return LDP_ERR;
	}

	/* Remove label from appropriate lists */
	switch (lg->type)
	{
		case LDP_LIB_TYPE_TRANSIT:
			lt = (ldp_lib_transit *)lg;
			/* Delete label from session list */
			rem_node(&lt->n_lss);
			break;
		//default:
			/* 
			 * Foreign entry. We are possibly
			 * called from foreign caller indicating
			 * this label is not needed.
			 */
	}

	/* Assume label is already removed from other lists and we can safely remove it */
	fib_delete(&lib->f, lg);

	/* Save label to free list for faster allocation */
	LOG_DB_DB2("%d < %d < %d, current label is %d", pl->label_min, label, pl->label_max, lib->current_label);

	/* If label is outside of configured limit scope, do nothing */
	if ((label < pl->label_min) || (label > pl->label_max))
		return LDP_OK;

	/* Check if re're freeing last allocated label */
	if (label + 1 == lib->current_label)
	{
		--lib->current_label;
		return LDP_OK;
	}

	LOG_DB_DBG("Adding label %u to free list", label);
	/* Add label to free list */
	fl = mb_allocz(MAIN_POOL(pl), sizeof(ldp_lib_free_label));
	fl->label = label;
	add_tail(&lib->free_items, NODE fl);

	return LDP_OK;
}

ldp_lib_general *
ldp_lib_label_find(struct proto_ldp *pl, u32 label)
{
	ldp_lib_database *lib = pl->lib;

	return (ldp_lib_general *)fib_find(&lib->f, &label, 0);
}

/**
 * ldp_lib_flabel_add - Request label for foreign protocol
 * @pl: pointer to protocol instance
 * @fp: pointer to requesting protocol
 * @info: data description
 * @ltype: label type requested
 *
 * Returns allocated label number or 0
 */
u32
ldp_lib_flabel_add(struct proto_ldp *pl, struct proto *fp, service_relfunc relfunc, char *info, int ltype, int mpls_op)
{
	ldp_lib_general *lg;

	if ((lg = ldp_lib_label_add(pl, ltype, mpls_op)) == NULL)
		return 0;
		
	/* Fill in info  */
	bsnprintf(lg->info, sizeof(lg->info), "%s", info);
	lg->proto = fp;
	lg->relfunc = relfunc;
	
	return lg->me.local_label;
}

int
ldp_lib_flabel_del(struct proto_ldp *pl, u32 label)
{
	return ldp_lib_label_del(pl, label);
}

/************************************************************
 *                                                          *
 *                    ADDRMAP DATABASE                      *
 *                                                          *
 ***********************************************************/
#define ADDRMAP_HASH_SIZE	65536

/*
 * Initialize nexthop mapping database
 */
int
ldp_addrmap_init(struct proto_ldp *pl)
{
	list *alist;
	int i;

	if ((alist = mb_alloc(MAIN_POOL(pl), sizeof(list) * ADDRMAP_HASH_SIZE)) == NULL)
	{
		LOG_ERR("Error allocating memory for addrmap");
		return LDP_ERR;
	}

	pl->addr_hash = alist;

	for (i = 0; i < ADDRMAP_HASH_SIZE; i++, alist++)
		init_list(alist);

	return LDP_OK;
}

/*
 * Shutdown addrmap database freeing all remaining nexthops
 */
void
ldp_addrmap_shutdown(struct proto_ldp *pl, int last UNUSED)
{
	list *alist;
	int i;
	ldp_addrmap *la;

	for (i = 0, alist = pl->addr_hash; i < ADDRMAP_HASH_SIZE; i++, alist++)
	{
		WALK_LIST_FIRST(la, *alist)
		{
			rem_node(NODE la);
			rem_node(&la->n_lss);
			mb_free(la);
		}
	}


	mb_free(pl->addr_hash);
	pl->addr_hash = NULL;
}


/*
 * Finds address mapping structure by nexthop IP address
 */
ldp_addrmap *
ldp_addrmap_find(struct proto_ldp *pl, ip_addr addr)
{
	ldp_addrmap *la;

	WALK_LIST(la, pl->addr_hash[ldp_addrmap_hash(addr)])
	{
		if (ipa_equal(la->addr, addr))
			return la;
	}

	return NULL;
}

int
ldp_addrmap_add(struct proto_ldp *pl, ldp_session *ls, ip_addr addr)
{
	ldp_addrmap *la;
	
	if ((la = mb_alloc(MAIN_POOL(pl), sizeof(ldp_addrmap))) == NULL)
		return LDP_ERR;

	la->ls = ls;
	la->addr = addr;
	/* Add to hash */
	add_tail(&pl->addr_hash[ldp_addrmap_hash(la->addr)], NODE la);
	/* Add to per-session list */
	add_tail(&ls->addr_list, &la->n_lss);

	return LDP_OK;
}

int
ldp_addrmap_del(struct proto_ldp *pl, ldp_session *ls, ip_addr addr)
{
	ldp_addrmap *la;

	if (((la = ldp_addrmap_find(pl, addr)) != NULL) && (la->ls == ls))
	{
		rem_node(NODE la);
		rem_node(&la->n_lss);
		mb_free(la);
		return LDP_OK;
	}

	return LDP_ERR;
}

void
ldp_addrmap_flush_session(struct proto_ldp *pl, ldp_session *ls)
{
	ldp_addrmap *la;

	WALK_LIST_FIRST(la, ls->addr_list)
	{
		la = SKIP_BACK(ldp_addrmap, n_lss, la);
		rem_node(NODE la);
		rem_node(&la->n_lss);
		mb_free(la);
	}
}

inline int 
ldp_addrmap_hash(ip_addr addr)
{
	return ipa_hash(addr);
}

char *
ldp_mplsop_texttype(enum kmpls_op op)
{
	char *buf;
	static char xbuf[50];

	switch (op)
	{
		case KMPLS_OP_NONE:
			buf = "NONE";
			break;

		case KMPLS_OP_POP:
			buf = "POP";
			break;

		case KMPLS_OP_PUSH:
			buf = "PUSH";
			break;

		case KMPLS_OP_SWAP:
			buf = "SWAP";
			break;

		case KMPLS_OP_L3VPN:
			buf = "L3VPN";
			break;

		case KMPLS_OP_NETGRAPH:
			buf = "NETGRAPH";
			break;

		default:
			bsnprintf(xbuf, sizeof(xbuf), "UNKN:%d", op);
			buf = xbuf;
	}

	return buf;
}

char *
ldp_lib_texttype(enum LDP_LIB_TYPE type)
{
	char *buf;
	static char xbuf[50];

	switch (type)
	{
		case LDP_LIB_TYPE_GENERAL:
			buf = "general";
			break;

		case LDP_LIB_TYPE_L3VPN:
			buf = "l3vpn";
			break;

		case LDP_LIB_TYPE_L2P2P:
			buf = "l2p2p";
			break;

		case LDP_LIB_TYPE_TRANSIT:
			buf = "swap";
			break;

		default:
			snprintf(xbuf, sizeof(xbuf), "unkn:%d", type);
			buf = xbuf;
	}

	return buf;
}
