/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/**
 * DOC: LDP
 *
 * LDP is a protocol defined for distributing labels. It is the set of
 * procedures and messages by which Label Switched Routers (LSRs)
 * establish Label Switched Paths (LSPs) through a network by mapping
 * network-layer routing information directly to data-link layer
 * switched paths.  These LSPs may have an endpoint at a directly
 * attached neighbor (comparable to IP hop-by-hop forwarding), or may
 * have an endpoint at a network egress node, enabling switching via all
 * intermediary nodes. This implementation is based on RFC5074.
 *
 */

/*
 * CODING STYLE:
 * Style derived from FreeBSD style(9) with minor excaptions:
 * - 'if {' is usually written as
 *   if
 *   {
 *
 *
 * NAMING SCHEME:
 *
 * Function naming:
 * 'Globally' (e.g. between protocol files) visible functions names by:
 * ldp[_subsystem]_action_info. Examples:
 * ldp_session_check_addrs, ldp_tlv_fill_chello_params
 * 'Locally' visible (single file) use the same scheme and (in some caces) 
 * can be prefixed with underscore
 *
 * Some exceptions:
 * * protocol hooks
 * * _texttype functions (converting various enums to human-readable form)
 *
 *
 * Structure naming:
 * 'Globally' visible structure is named by
 * _ldp_name. Evey such structure should have a typedef to ldp_name
 *
 * Exception:
 * * proto_ldp structure
 *
 * Common variables naming:
 * struct protocol: p
 * struct proto_ldp: pl
 * ldp_interface: li
 * ...
 *
 * Function grouping:
 * externally visible functions first,
 * local - next
 * hooks - last
 *
 * FILE STRUCTURE:
 *
 * Globally visible functions should come first,
 * locally visible follows,
 * vairous hooks comes last
 *
 * General notes:
 * WALK_LIST_DELSAFE uses (xx, xx_next) variable naming
 *
 * DO NOT USE WALK_LIST. Really. Copy&Paste it to the place with SKIP_BACK and..
 *
 * If we free some structure field we set it to NULL
 *
 * EVERY hook (protocol, timer, socket, ...) SHOULD
 * have
 * DEBUG_PROTO(pl) in the beginning (if it is possible to assign protocol instance)
 * and 
 * UNDEBUG_PROTO() in the end.
 * This is needed for debugging stuff to work
 *
 * 
 *
 * MEMORY MANAGEMENT:
 * Items we allocate from our base pool (using MAIN_POOL(pl) macro):
 * * every long-life items (timers, hashes, sockets, other general items)
 * * LDP objects (hello adjacencies, neighbours, sessions)
 *
 * Items we allocate from PDU pool:
 * * PDUs/message queue items (if stack/linpool allocation is not possible)
 *
 * Items we allocate on pdu_lp linpool/stack if possible:
 * Every data needed for parsing received PDU
 *
 */

/*
 * General protocol setup and most external world callbacks (routing table
 * updates, interface updates, reconfiguration, global timers) are
 * located in this file.
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_proto.h"
#include "nest/cli.h"

list	pl_list; /* list of LDP protocols */
int	init_done;

int ldp_dispatch_service_request(struct proto *p, struct service_request *req, byte **buf, size_t *buflen);

/* General protocol routines */

/*
 * Starts protocol instance.
 */
static int
ldp_start(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_config *c = (void *) p->cf;

	DEBUG_PROTO(pl);

	/* Copy debugging mask at first */
	memcpy(&pl->debug_mask, &c->debug_mask, sizeof(pl->debug_mask));

	if ((pl->pdu_pool = rp_new(p->pool, "LDP PDU pool")) == NULL)
	{
		LLOG_ERR("Error creating LDP PDU pool");
		goto error;
	}

	/* Setup linear pool for structure parsing */
	if ((pl->pdu_lp = lp_new(p->pool, 1024)) == NULL)
	{
		LLOG_ERR("Error creating LDP linear pool");
		goto error;
	}

	/* Setup routing/forwarding database */
	if ((ldp_rib_init(pl) != LDP_OK) || (ldp_lib_init(pl) != LDP_OK) || (ldp_addrmap_init(pl) != LDP_OK))
	{
		LLOG_ERR("Error initializing databases");
		goto error;
	}

	/* Setup global listening socket/our addresses/timers */
	if ((ldp_sessions_init(pl) != LDP_OK))
	{
		LLOG_ERR("Error initializing sessions module");
		return PS_STOP;
	}
	if ((ldp_hellos_init(pl) != LDP_OK))
	{
		LLOG_ERR("Error initializing hello module");
		return PS_STOP;
	}

	LP_COPY(c, label_min);
	LP_COPY(c, label_max);

	LOG_DBG("Start: LDP range: %d - %d", pl->label_min, pl->label_max);

	pl->router_id = config->router_id;
	/* mpls_table is mandatory config attribute */
	pl->mpls_table = c->mpls_table->table;
	rt_lock_table(pl->mpls_table);
	LOG_DBG("Router ID: %R, MPLS table %s", pl->router_id, pl->mpls_table->name);

	/* Configure protocol as service provider */
	proto_add_service_hook(p, SRCLASS_LABEL, (service_func)&ldp_dispatch_service_request);

	ldp_ifaces_reconfigure(pl, c);

	ldp_session_notify_reconfigure(pl);

	add_tail(&pl_list, &pl->pl_list);

	UNDEBUG_PROTO();
	return PS_UP;

error:
	/* Free various resources */
	lp_flush(pl->pdu_lp);
	pl->pdu_lp = NULL;
	rfree(pl->pdu_pool);
	pl->pdu_pool = NULL;

	UNDEBUG_PROTO();
	return PS_STOP;
}

/*
 * Shuts protocol down to unconfigured state
 */
static int
ldp_shutdown(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	int last;
	ldp_neighbour *ln, *ln_next;

	DEBUG_PROTO(pl);

	rt_unlock_table(pl->mpls_table);
	pl->mpls_table = NULL;

	/* Remove protocol from LDP protocols list */
	rem_node(&pl->pl_list);
	last = EMPTY_LIST(pl_list) ? 1 : 0;

	/* Shut down neighbours first to permit notifications */
	WALK_LIST_DELSAFE(ln, ln_next, pl->neighbour_list)
	{
		ln = SKIP_BACK(ldp_neighbour, n_list, ln);
		ldp_neighbour_shutdown(ln);
	}

	/* Stop discovery timers and kill remaining hellos if any */
	ldp_hellos_shutdown(pl, last);
	
	/* Close remaining sessions, kill socket and timers */
	ldp_sessions_shutdown(pl, last);

	/* 
	 * Free all interfaces.
	 * This should be called AFTER ldp_hellos_shutdown()
	 * to ensure no hellos holding reference to interfaces
	 */
	ldp_ifaces_shutdown(pl);

	/* Free RIB/LFIB/addrmap databases */
	ldp_rib_free(pl, last);
	ldp_lib_free(pl, last);
	ldp_addrmap_shutdown(pl, last);

	/* Free resource pools */
	lp_flush(pl->pdu_lp);
	pl->pdu_lp = NULL;
	rfree(pl->pdu_pool);
	pl->pdu_pool = NULL;

	pl->tcp_saddr = IPA_NONE;

	LOG_DBG("Going down..");
	UNDEBUG_PROTO();


	return PS_DOWN;
}

/*
 * Does protocol reconfiguration.
 * Returns 1 if reconfiguration succeeded or
 * 0 if soft reconfiguration can'te be done
 */
static int
ldp_reconfigure(struct proto *p, struct proto_config *c)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_config *old = (struct ldp_config *) (p->cf);
	struct ldp_config *new = (struct ldp_config *) c;

	DEBUG_PROTO(pl);

	LOG_VRB("Protocol reconfiguration requested");

	if (old->distribute_mode != new->distribute_mode) {
		LOG_ERR("Distribute mode cannot be changed on fly");
		goto error;
	}

	if (old->retention_mode != new->retention_mode) {
		LOG_ERR("Retention mode cannot be changed on fly");
		goto error;
	}

	if (old->lspace_type != new->lspace_type) {
		LOG_ERR("Label space type cannot be changed on fly");
		goto error;
	}

	if (old->label_min < new->label_min) {
		/* 
		 * If we have LDP neighbours to which we are advertising some labels
		 * low label numbers are definitely used, so the only way is to
		 * shut down this instance and start new. If we have no LDP neighbours
		 * restarting protocol does no harm, so we simply return 0 here.
		 */
		LOG_ERR("Minimum label value cannot be increased due to label usage");
		goto error;
	}

	if ((old->label_max < new->label_max) && (1 == 2)) {
		/*
		 * A bit more tricky situation, we should check current label usage somehow
		 * FIXME: implement some request to LFIB
		 */
		LOG_ERR("Maximum label value cannot be decreased due to label usage");
		goto error;
	}

	/* Checks completed, new configuration can be applied on fly */
	LP_COPY(new, label_min);
	LP_COPY(new, label_max);
	memcpy(&pl->debug_mask, &new->debug_mask, sizeof(pl->debug_mask));

	/* Reconfigure all participating interfaces */
	ldp_ifaces_reconfigure(pl, new);

	/* Notify sessions aboute some re-configuration taking place */
	ldp_session_notify_reconfigure(pl);

	return 1;
error:
	LOG_VRB("Protocol soft reconfiguration failed");
	UNDEBUG_PROTO();
	return 0;
}

int
ldp_dispatch_service_request(struct proto *p, struct service_request *req, byte **buf, size_t *buflen)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_label_request *lreq;
	struct ldp_label_release *lrel;
	struct service_reply *repl;
	int directed = req->target ? 1 : 0;
	int error = SR_FAIL;
	u32 label, *plabel;

	DEBUG_PROTO(pl);

	LOG_DBG("Service request: %d/%d from protocol %s", req->request, req->subclass, p->name);

	if (req->request != SRCLASS_LABEL)
	{
		error = directed ? SR_FAIL : SR_NEXT;
		goto done;
	}

	if (req->subclass != SRSUB_LABEL_REQ)
	{
		error = directed ? SR_FAIL : SR_NEXT;
		goto done;
	}

	switch (req->subclass)
	{
		case SRSUB_LABEL_REQ:
			/* Dispatch label request */

			LOG_DBG("subclass: Label request");
		
			/* work with supplied buffer only */
			if (*buf == NULL)
				break;
		
			lreq = (struct ldp_label_request *)req->data;
			if ((label = ldp_lib_flabel_add(pl, req->p, req->relfunc, lreq->description, lreq->ltype, KMPLS_OP_NETGRAPH)) == 0)
				break;
			/* Fill in reply */
			repl = (struct service_reply *)(*buf);
			repl->p = &pl->proto;
			/* Return label to the caller */
			plabel = (u32 *)repl->data;
			*plabel = label;
			*buflen = sizeof(struct service_reply) + sizeof(u32);

			LOG_DBG("Allocated label %u", label);
		
			error = SR_OK;
			break;

		case SRSUB_LABEL_REL:
			/* Release local label */
			lrel = (struct ldp_label_release *)req->data;
			LOG_DBG("subclass: Label delete request for label %u", lrel->label);
			error = (ldp_lib_flabel_del(pl, lrel->label) == LDP_OK) ? SR_OK : SR_FAIL;
			break;

		default:
			/* Unknown subclass */
			error = directed ? SR_FAIL : SR_NEXT;
			break;
	}
done:
	UNDEBUG_PROTO();
	return error;
}

/*
 * Dumps protocol state on debugging output.
 * A bit outdated
 *
 */
static void
ldp_dump(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_iface *li;

	DEBUG_PROTO(pl);

	debug("Interfaces:\n");
	WALK_LIST(li, pl->iface_list)
	{
		debug("Interface: %s\n", li->iface->name);
		debug("LDP address: %I/%d\n", li->addr->prefix, li->addr->pxlen);
		debug("LDP status: %s\n", ldp_iface_textstatus(li->status));
	}

	UNDEBUG_PROTO();
}

/*
 * Interface state change notification.
 * Pass to iface layer.
 * Generally we update our local states and
 * checking if our local address is changed
 */
static void
ldp_if_notify(struct proto *p, unsigned flags, struct iface *i)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;

	DEBUG_PROTO(pl);
	LLOG_DBG("Called on interface %s, flags 0x%X", i->name, flags);
	ldp_iface_update_state(pl, i, flags);
	UNDEBUG_PROTO();
}

/*
 * Interface address change notification.
 * Pass to iface layer.
 */
void
ldp_ifa_notify(struct proto *p, unsigned flags, struct ifa *a)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;

	DEBUG_PROTO(p);
	LLOG_DBG("Called on interface %s, flags 0x%X", a->iface->name, flags);
	ldp_iface_update_addrs(pl, a, flags);
	UNDEBUG_PROTO();
}

/*
 * Checks parsed LDP config
 */
void
ldp_check(struct ldp_config *c)
{
	if (c->mpls_table == NULL)
		cf_error("Please set mpls_table for an LDP protocol instance");
}

/*
 * Create new protocol instance. initialize its callbacs/lists
 */
static struct proto *
ldp_init(struct proto_config *c)
{
	struct proto *p;
	struct proto_ldp *pl;

	/* Init protocol list */
	if (init_done == 0)
	{
		init_done = 1;
		init_list(&pl_list);
	}

	p = proto_new(c, sizeof(struct proto_ldp));
	pl = (struct proto_ldp *) p;
	DEBUG_PROTO(pl);

	//p->neigh_notify = ldp_neigh_notify;
	p->if_notify = ldp_if_notify;
	p->ifa_notify = ldp_ifa_notify;
	p->accept_ra_types = RA_PRE;
	//p->rt_notify = ldp_rt_notify;
	p->rt_notify_pre = ldp_rt_notify_pre;

	ldp_ifaces_init(pl);

	UNDEBUG_PROTO();

	return p;
}

struct protocol proto_ldp = {
  name:		"LDP",
  template:	"ldp%d",
  attr_class:	EAP_LDP,
  init:		ldp_init,
  dump:		ldp_dump,
  start:	ldp_start,
  shutdown:	ldp_shutdown,
  reconfigure:	ldp_reconfigure,
  get_attr:	ldp_get_attr,
  get_route_info:	ldp_get_route_info
};

