/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

#ifndef _BIRD_LDP_H_
#define _BIRD_LDP_H_

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>

#include "nest/bird.h"
#include "nest/iface.h"
#include "nest/protocol.h"
#include "nest/route.h"
#include "conf/conf.h"
#include "lib/string.h"
#include "lib/alloca.h"
#include "lib/socket.h"
#include "ldp_debug.h"

#define LDP_LABEL_MIN	16
#define LDP_LABEL_MAX	0xFFFFF

#define LDP_SPACE_GLOBAL	0x01 /* Global pattern space, */
#define	LDP_SPACE_INTERFACE	0x02 /* */

/* RFC 5036 DEFINITIONS */
#define LDP_LABEL_SPACE_GLOBAL		0x01
#define LDP_LABEL_SPACE_INTERFACE	0x02
/* Label Distribution Control Mode, clause 2.6.1 */
#define LDP_DISTRIBUTE_INDEPENDENT	0x01
#define LDP_DISTRIBUTE_ORDERED		0x02
/* Label Retention Mode, clause 2.6.2 */
#define LDP_RETENTION_CONSERVATIVE	0x01
#define LDP_RETENTION_LIBERAL		0x02
/* Label Advertisement Mode, clause 2.6.3 */
#define LDP_ADVERTISE_UNSOLICITED	0x01
#define LDP_ADVERTISE_ONDEMAND		0x02

/* Interface types definition */
#define LDP_IF_MULTICAST		0x01 /* Interface supports multicast (IFF_MULTICAST set) */
#define LDP_IF_NBMA			0x02 /* All other interfaces */

/* Common flags shared within all LDP structures */
#define LDP_FLAG_DELETED		0x0001
#define LDP_FLAG_DONTFREE		0x0002
#define LDP_FLAG_HOLDTIMER		0x0004 /* Hold timer value is specific to interface */
#define LDP_FLAG_KEEPTIMER		0x0004 /* Hold timer value is specific to interface */
#define LDP_FLAG_SOCKETUP		0x0008 /* LDP socket is set up on interface */
#define LDP_FLAG_HELLOPDU		0x0010 /* UDP Hello PDU */
#define LDP_FLAG_SESSION_ACTIVE		0x0100 /* Session (Neigbour really) mode is active */
#define LDP_FLAG_CUSTOM_BUF		0x0200 /* Object has custom databuf that has to be freed separately */
#define LDP_FLAG_DELETED2		0x0400 /* Used in transport session destroying */
#define LDP_FLAG_TRANSMIT_RUNNING	0x0800 /* Indicates that data transmit on socket is running */
#define LDP_FLAG_INSTALL_RIB		0x1000 /* used in ldp_session_install_prefix */
#define LDP_FLAG_INSTALL_LFIB		0x2000 /* used in ldp_session_install_prefix */
#define LDP_FLAG_ANNOUNCE		0x4000 /* used in ldp_session_install_prefix */
/* Message sending flags */
#define LM_NONE				0x00	/* No flags */
#define LM_IMMEDIATE			0x01	/* Request immediate sending */
#define LM_BULK				0x02	/* Message contains multiple messages */
#define LM_NOFREE			0x04	/* Do not free supplied message buffer */

#define _IS_DELETED(x)	((x)->flags & LDP_FLAG_DELETED)
#define _SET_DELETED(x)	((x)->flags |= LDP_FLAG_DELETED)

#define _IS_DELETED2(x)		((x)->flags & LDP_FLAG_DELETED2)
#define _SET_DELETED2(x)	((x)->flags |= LDP_FLAG_DELETED2)
#define _UNSET_DELETED2(x)	((x)->flags &= ~LDP_FLAG_DELETED2)

#define LDP_HASH_SIZE			65536	/* neighbour, session and hello hash size */

#define LDP_SOCK_RX			0x01
#define LDP_SOCK_TX			0x02
#define LDP_SOCK_RX_BUFFER		65536
#define LDP_SOCK_TX_BUFFER		65536


#define LDP_OBJ_OK			0	/* TLV/Message parsed successfully */
#define LDP_OBJ_FAIL			1	/* Object is unknown and unknown bit is clear */
#define LDP_OBJ_IGNORE			2	/* Object is unknown and unknown bit is set */
#define LDP_OBJ_FORWARD			3	/* Object is unknown and unknown|forward bits are set */
#define LDP_OBJ_FATAL			4	/* Object causes protocol fatal error */

/* RFC5036, Clause 3.1 */
#define LDP_MAXPDU_INITIAL		4096

#define LDP_HELLO_BASE_BUFFER		LDP_MAXPDU_INITIAL * 2
#define LDP_SESSION_BASE_BUFFER		LDP_MAXPDU_INITIAL * 2

/* PDU TYPES */
#define LDP_PDU_HELLO		1
#define LDP_PDU_SESSION		2

/* 
 * Route attributes we use.
 * Some attributes can reside in ldp_ext.h
 */
#define	LA_MPLSENTRY		0	/* LA_MPLSENTRY */

/*
 * Request subclasses we use
 */
/* SRCLASS_LABEL */
#define SRSUB_LABEL_REQ		1
struct ldp_label_request
{
	int ltype;
	char *description;
};
#define SRSUB_LABEL_REL	2
struct ldp_label_release
{
	u32 label;
};

/* Interface status FSM */
enum LDP_ISM {
	LDP_ISM_DOWN,
	LDP_ISM_INIT,
	LDP_ISM_UP,
	LDP_ISM_INVALID
};

/* Session state FSM */
enum LDP_SSM {
	LDP_SSM_NONEXISTENT,
	LDP_SSM_INITIALIZED,
	LDP_SSM_OPENSENT,
	LDP_SSM_OPENREC,
	LDP_SSM_OPERATIONAL
};

enum LDP_SESS_ROLE {
	LDP_SESSION_PASSIVE,
	LDP_SESSION_ACTIVE
};

enum LDP_SESS_TR {
	LDP_SESSION_TRANSPORT_NONE,
	LDP_SESSION_TRANSPORT_INIT,
	LDP_SESSION_TRANSPORT_ACTIVE
};

typedef struct _ldp_neighbour {
	node		n_hash;		/* Per hash value List */
	node		n_list;		/* Global list with all neighbours */
	uint32_t	lsr_id;		/* Neigbour id */
	int		flags;		/* Per-neighbour flags */
	sock		*sk;		/* TCP socket for transport LDP session */
	enum LDP_SESS_TR	tr_state;	/* LDP Transport session state */
	ip_addr		local;		/* Local address used in TCP session */
	ip_addr		remote;		/* Remote address used in TCP session */
	time_t		last_packet;	/* Last time packet is received */
	list		session_list;	/* List of logical LDP sessions */
	list		hello_list;	/* List of all hello adjacencies */
	list		addr_list;	/* List of all nexthops recorded for given neighbour */
	list		packets_list;	/* List of messages queued for sending */
	int		packets_length;	/* Total _message_ length queued for sending */
	struct proto_ldp	*pl;	/* Pointer to protocol structure */
	int		dataoff;	/* Current data offset in data buffer */
	int		datasize;	/* Data buffer size */
	char		*data;		/* Pointer for data buffer */
} ldp_neighbour;

typedef struct _ldp_session {
	node		n_hash;		/* Per hash value List */
	node		n_list;		/* Global list with all sessions */
	node		n_lsr;		/* Per-neighbour session list */
	int		flags;
	int		seq;		/* Session sequence number */
	int		max_pdu;	/* Consiled maximum PDU */
	uint32_t	lsr_id;		/* Copy of neighbour router address */
	uint16_t	space_id;	/* Remote advertized label space */
	enum LDP_SSM	state;		/* State machine current state */
	time_t		last_state;	/* Last time state is changed */
	time_t		last_received;	/* Last time packet is received */
	timer		*hold_timer;	/* KeepAlive timer */
	time_t		last_sent;	/* Last time packet is sent */
	timer		*send_timer;	/* KeepAlive send timer */
	int		keep_interval;	/* Session keepalive interval */
	int		messages_in;	/* Received message count */
	int		messages_out;	/* Transmit message count */
	struct ldp_iface	*li;	/* Interface session is established via, can be NULL */
	ldp_neighbour	*ln;		/* Pointer to LDP neighbour */
	list		addr_list;	/* List of all interface addresses reported by LSR */
	list		prefix_list;	/* List of all prefixes reported by LSR */
	list		label_list;	/* List of all labels allocated for prefixes */
} ldp_session;

typedef struct _ldp_qmessage {
	node		n;
	ldp_session	*ls;		/* Pointer to LDP session, used in MAXPDU calculations */
	int		length;		/* Message length */
	int		messages;	/* Total message count (used in bulk send) */
} ldp_qmessage;
#define LDP_QHEADER_SIZE	BIRD_ALIGN(sizeof(ldp_qmessage), CPU_STRUCT_ALIGN)
#define LDP_QHEADER(x)		((byte *)(x) + LDP_QHEADER_SIZE)
#define LDP_QDATA(x)		(byte *)(LDP_QHEADER(x) + LDP_MESSAGE_HEADER_SIZE)

typedef struct _ldp_hello {
	node		n_hash;		/* Per hash value List */
	node		n_list;		/* Global list with all hello */
	node		n_lsr;		/* Per LSR list */
	int		flags;
	ip_addr		remote;		/* Address hello received from */
	ip_addr		transport;	/* Session transport address */
	uint16_t	space_id;	/* Remote advertized label space*/
	time_t		last_packet;	/* Last time packet is received */
	timer		*hold_timer;	/* Hold down timer */
	int		dataoff;	/* Current data offset in data buffer */
	int		datasize;	/* Data buffer size */
	struct ldp_iface	*li;	/* Interface hello is received via */
	ldp_neighbour	*ln;		/* Neihbour associated with this hello */
	byte		*data; 		/* Pointer for data buffer */
} ldp_hello;


struct ldp_config {
	struct			proto_config c;
	int			label_min;	/* Minimum LDP label to start with */
	int			label_max;	/* Maximum LDP label to end with */
	int			distribute_mode;
	int			retention_mode;
	u8			debug_mask[LD_MAX];	/* Severity debug masks */
	int			lspace_type;	/* LDP database type */
	int			loop_detection;	/* LDP Loop detection switch */
	int			dhold_interval;	/* Discovery hold interval */
	int			dsend_interval;	/* Discovery sending interval */
	int			thold_interval;	/* Targeted hold interval */
	int			tsend_interval;	/* Targeted sending interval */
	int			skeep_interval;	/* Session KeepAlive interval */
	list			if_patt_list;	/* Interface patterns list */
	struct rtable_config	*mpls_table;	/* Table used for storing labels */
};

struct proto_ldp {
	struct proto		proto;
	uint32_t		router_id;	/* LDP router id */
	int			space_id;	/* LDP labelspace ID */
	int			label_min;	/* Minimum LDP label to start with */
	int			label_max;	/* Maximum LDP label to end with */
	int			distribute_mode;
	int			retention_mode;
	u8			debug_mask[LD_MAX];	/* Severity debug masks */
	rtable			*mpls_table;	/* Table used for storing labels */
	node			pl_list;	/* list of LDP protocols */
	int			cfg_index;	/* Configuration counter */
	int			lspace_type;	/* LDP database type */
	int			loop_detection;	/* LDP Loop detection switch */
	linpool			*pdu_lp;	/* Linear pool for storing current PDUs with data */
	pool			*pdu_pool;	/* Pool for storing queued PDUs */
	sock			*tcp_sk;	/* TCP socket for binding per-interface IP address */
	ip_addr			tcp_saddr;	/* Source address announced in LDP hellos */
	timer			*dhello_timer;	/* Discovery Hello timer */
	timer			*thello_timer;	/* Targeted Hello timer */
	timer			*xmit_timer;	/* Periodic transmit timer */
	list			*neighbour_hash;	/* dynamically allocated neighbour hash */
	list			*session_hash;	/* dynamically allocated session hash */
	list			*hello_hash;	/* dynamically allocated neighbour hash */
	list			*addr_hash;	/* dynamically allocated nexhop database */
	/* The following lists contains data for faster show/delete commands */
	list			neighbour_list;	/* All neighbours seen */
	list			session_list;	/* All neighbour sessions */
	list			hello_list;	/* All incomplete hello list */
	list			iface_list;	/* Enabled interfaces */
	struct _ldp_lib_database	*lib;	/* Label database (LIB) */
	struct fib		rib;		/* Lmap database (RIB) */
};
#define MAIN_POOL(x)		((struct proto *)(x))->pool

struct ldp_iface {
	node			n;
	int			flags;
	struct iface		*iface;		/* Pointer to nest iface */
	struct ifa		*addr;		/* IP prefix associated with that LDP iface */
	struct proto_ldp	*pl;		/* Pointer to parent protocol */
	sock			*rx_sk;		/* IP socket (for multicast receive) */
	sock			*tx_sk;		/* IP socket (for multicast transmit) */
	int			type;		/* Interface type: Broadcast, NBMA */
	enum LDP_ISM		status;		/* Interface state machine status */
	int			refcount;	/* Object reference count */
	int			advertise_mode;
	int			hold_timer;	/* Value for hold timer */
	int			keepalive_timer;	/* Value for keepalive timer */
};
#define iface_name		iface->name

struct ldp_iface_patt {
	struct iface_patt	i;
	int			flags;
	int			advertise_mode;
	int			hold_timer;	/* Value for hold timer */
	int			keepalive_timer;	/* Value for keepalive timer */
};

struct _ldp_addrmap;
struct _ldp_rib_entry;
struct _ldp_lib_general;

/* ldp.c */
extern list pl_list;
void ldp_check(struct ldp_config *c);

/* cli.c */
int ldp_get_attr(eattr *a, byte *buf, int buflen);
void ldp_get_route_info(rte *rte, byte *buf, ea_list *attrs);
void ldp_show_neighbour(struct proto *p);
void ldp_show_forwardingtable(struct proto *p);
void ldp_show_bindings(struct proto *p);
void ldp_show_interfaces(struct proto *p);

/* route.c */
void ldp_rt_notify_pre(struct proto *p, rtable *tbl UNUSED, net *n, rte *new, rte *old, ea_list *attrs);
void ldp_rt_update(struct proto_ldp *pl, net *n, rte *old, struct _ldp_rib_entry *re);
void ldp_rt_announce_label(struct proto_ldp *pl, struct _ldp_lib_general *lg, ip_addr nexthop, ldp_session *ls);
void ldp_rt_withdraw_label(struct proto_ldp *pl, bird_mpls_entry *me);

/* iface.c */
void ldp_iface_update_state(struct proto_ldp *pl, struct iface *i, unsigned flags);
void ldp_iface_update_addrs(struct proto_ldp *pl, struct ifa *a, unsigned flags);
void ldp_ifaces_reconfigure(struct proto_ldp *pl, struct ldp_config *c);
void ldp_ifaces_shutdown(struct proto_ldp *pl);
int ldp_ifaces_init(struct proto_ldp *pl);
char *ldp_iface_textstatus(enum LDP_ISM status);

/* session.c */
int ldp_sessions_init(struct proto_ldp *pl);
void ldp_sessions_shutdown(struct proto_ldp *pl, int last);
void ldp_session_notify_reconfigure(struct proto_ldp *pl);
char *ldp_session_textstatus(enum LDP_SSM state);

ldp_session *ldp_session_add(ldp_hello *lh, ip_addr remote);
void ldp_session_shutdown(ldp_session *ls);
ldp_session *ldp_session_find(struct proto_ldp *pl, uint32_t lsr_id, uint16_t space_id);
void ldp_session_prepare(ldp_session *ls);
int ldp_session_init_connect(ldp_session *ls);
enum LDP_SSM ldp_session_set_state(ldp_session *ls, enum LDP_SSM state);

ip_addr ldp_session_global_saddr(struct proto_ldp *pl);

ldp_neighbour *ldp_neighbour_find(struct proto_ldp *pl, uint32_t addr);
ldp_neighbour *ldp_neighbour_add(struct proto_ldp *pl, uint32_t lsr_id);
void ldp_neighbour_shutdown(ldp_neighbour *ln);

int ldp_session_install_prefix(struct proto_ldp *pl, ldp_session *ls, struct _ldp_rib_entry *re, net *n, rte *e, int flags);
ldp_qmessage *ldp_session_get_buffer(ldp_session *ls, int *len);
void ldp_session_announce_lmap(struct proto_ldp *pl, struct _ldp_rib_entry *re, ldp_session *one, int flags, int withdraw);
int ldp_session_sendto_all_butone(struct proto_ldp *pl, ldp_qmessage *lq, ldp_session *one, int flags);
int ldp_session_send_message(ldp_session *ls, ldp_qmessage *lq, int flags);

/* ldp_hello.c */
int ldp_hellos_init(struct proto_ldp *pl);
void ldp_hellos_shutdown(struct proto_ldp *pl, int last UNUSED);
int ldp_setup_hello_socket(struct ldp_iface *li, int direction);
int ldp_hello_build(byte *databuf, int datasize, struct proto_ldp *pl);
struct _ldp_hello * ldp_hello_add(struct ldp_iface *li, ip_addr remote);
void ldp_hello_shutdown(ldp_hello *lh);

/* ldp_database.c */
char * ldp_mplsop_texttype(enum kmpls_op op);
/* addrmap */
int ldp_addrmap_init(struct proto_ldp *pl);
void ldp_addrmap_shutdown(struct proto_ldp *pl, int last UNUSED);
struct _ldp_addrmap * ldp_addrmap_find(struct proto_ldp *pl, ip_addr addr);
int ldp_addrmap_add(struct proto_ldp *pl, ldp_session *ls, ip_addr addr);
int ldp_addrmap_del(struct proto_ldp *pl, ldp_session *ls, ip_addr addr);
void ldp_addrmap_flush_session(struct proto_ldp *pl, ldp_session *ls);
/* LFIB */
int ldp_lib_init(struct proto_ldp *pl);
void ldp_lib_free(struct proto_ldp *pl, int last UNUSED);
u32 ldp_lib_flabel_add(struct proto_ldp *pl, struct proto *fp, service_relfunc relfunc, char *info, int ltype, int mpls_op);
int ldp_lib_flabel_del(struct proto_ldp *pl, u32 label);
/* RIB */
int ldp_rib_init(struct proto_ldp *pl);
void ldp_rib_free(struct proto_ldp *pl, int last UNUSED);

struct _ldp_pdu_header;
/* ldp_packet.c */
char * ldp_message_texttype(int type);
char * ldp_label_texttype(int label);
int ldp_pdu_preprocess(struct proto_ldp *pl, ip_addr remote, int maxlength, char *sock_data, int sock_size, char *prev_data, int prev_size, int prev_off, int *read_size);
void ldp_pdu_get_header(byte *data, struct _ldp_pdu_header *lph);
int ldp_pdu_put_header(byte *data, int length, int space_id);
void ldp_pdu_dispatch(ldp_neighbour *ln, byte *data, int length, int type, void *xptr);
int ldp_message_put_header(byte *data, ldp_session *ls, int unknown, int type, int length, int *id);

/* ldp_tlv.c */
struct _ldp_tlv;
struct _ldp_tlv_header;
struct _ldp_tlv_chello_params;
struct _ldp_tlv_csess_params;
struct _ldp_tlv_ipv4transport;
struct _ldp_tlv_addrlist;
struct _ldp_tlv_fec_element;
struct _ldp_tlv_generic_label;

void ldp_tlv_put_header(byte *data, int unknown, int forward, int type, int length);
void ldp_tlv_get_header(byte *data, struct _ldp_tlv_header *lth);
int ldp_tlv_put_ipv4transport(byte *data, struct proto_ldp *pl);
int ldp_tlv_put_chello_params(byte *data, struct proto_ldp *pl, int targeted);
int ldp_tlv_put_csess_params(byte *data, ldp_session *ls);
int ldp_tlv_put_status_params(byte *data, ldp_session *ls);
int ldp_tlv_fill_iaddrs(int af, byte *data, int datalen, int *_total);
int ldp_tlv_put_addrlist_initial(byte *data, size_t datasize, int af, byte *addrbuf, size_t addrbufsize, size_t *bufused);
int ldp_tlv_put_addrlist_single(byte *data, size_t datasize, int af, ip_addr addr);
int ldp_tlv_fec_put_prefix(byte *data, int datasize, ip_addr addr, int mask);
int ldp_tlv_put_fec(byte *data, int length);
int ldp_tlv_put_generic_label(byte *data, size_t datasize, int label);
int ldp_tlv_parse(struct proto_ldp *pl, struct _ldp_tlv_header *lth, byte *data, int length, struct _ldp_tlv **tlv);
struct _ldp_tlv * _ldp_tlv_find(list *l, int type);
#define ldp_tlv_find(l, type, cast)	(cast *)_ldp_tlv_find(l, type)
char * ldp_tlv_texttype(int type);

/* ldp_message.c */
struct _ldp_message_header;
int ldp_session_send_init(ldp_session *ls);
int ldp_session_send_keepalive(ldp_session *ls);
int ldp_session_send_addrlist_initial(ldp_session *ls);
int ldp_session_make_lmap(byte *data, size_t datasize, ldp_session *ls, ip_addr prefix, int plen, int label, int withdraw);
int ldp_dispatch_hello(ldp_neighbour *ln, ldp_hello *lh, struct _ldp_message_header *lmh, list *tlv_list);
int ldp_dispatch_session_init(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list);
int ldp_dispatch_session_keepalive(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list);
int ldp_dispatch_session_addrlist(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list);
int ldp_dispatch_session_lmap(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list, list *forward_list);
int ldp_dispatch_session_lreq(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list, list *forward_list);
int ldp_dispatch_session_lrel(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list, list *forward_list);
int ldp_dispatch_session_notify(ldp_neighbour *ln, ldp_session *ls, struct _ldp_message_header *lmh, list *tlv_list, list *forward_list);


/* better than nothing */
#define CurrentTime	time(NULL)


#define _ZERO(x)	memset(&(x), 0, sizeof(x))
#define _eq(x, y)	(!strcmp(x, y))
#define PL_TO_CF(x)	((struct ldp_config *)((struct proto *)(x))->cf)

#define _HOSTL(x)	(x) = ntohl(x)
#define _HOSTS(x)	(x) = ntohs(x)
#define _NETL(x)	(x) = htonl(x)
#define _NETS(x)	(x) = htons(x)

#define	LDP_OK		1
#define LDP_ERR		0

#define LDP_REF(x)	((x)->refcount++)
#define LDP_UNREF(x)	((x)->refcount--)

#define _LP_COPY(_pl, _c, _field)	(_pl)->_field = (_c)->_field
#define LP_COPY(_c, _field)	_LP_COPY(pl, _c, _field)

#define _LI_COPY(_li, _lip, _field)	(_li)->_field = (_lip)->_field
#define LI_COPY(_field)		_LI_COPY(li, lip, _field)

#endif
