/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * High-level LDP messages dispatch/generation routines
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_proto.h"
#include "ldp_database.h"


static int ldp_dispatch_session_common(ldp_session *ls, ldp_message_header *lmh);
static int ldp_session_send_lmap_initial(ldp_session *ls);

int
ldp_session_send_init(ldp_session *ls)
{
	int length, buflen = 0;
	ldp_qmessage *lq;
	
	if ((lq = ldp_session_get_buffer(ls, &buflen)) == NULL)
		return LDP_ERR;

	/* Add mandatory hello options TLV */
	length = ldp_tlv_put_csess_params(LDP_QDATA(lq), ls);

	/* Fill in message header */
	length = ldp_message_put_header(LDP_QHEADER(lq), ls, LDP_BIT_NONE, LDP_MSG_INIT, length, NULL);

	ldp_session_send_message(ls, lq, LM_IMMEDIATE);

	return LDP_OK;
}

int
ldp_session_send_keepalive(ldp_session *ls)
{
	int length, buflen = 0;
	ldp_qmessage *lq;
	
	if ((lq = ldp_session_get_buffer(ls, &buflen)) == NULL)
		return LDP_ERR;

	length = ldp_message_put_header(LDP_QHEADER(lq), ls, LDP_BIT_NONE, LDP_MSG_KEEPALIVE, 0, NULL);
	ldp_session_send_message(ls, lq, LM_IMMEDIATE);

	return LDP_OK;
}

int
ldp_session_send_addrlist_initial(ldp_session *ls)
{
	int length, cur_length = 0, alloc = 0, buflen = 0;
	size_t bufused;
	struct proto_ldp *pl = ls->ln->pl;
	ldp_qmessage *lq;
	int af;
	byte abuf[8192]; /* Get up to 2048 IPv4 addresses */
	byte *aptr;

	af = AF_INET;
	aptr = abuf;

	if (ldp_tlv_fill_iaddrs(af, aptr, sizeof(abuf), &length) == LDP_ERR)
	{
		/* Allocate more memory from main pool */
		if ((aptr = mb_alloc(MAIN_POOL(pl), length)) == NULL)
			return LDP_ERR;

		/* XXX: hack not to change length if it get changed */
		ldp_tlv_fill_iaddrs(af, aptr, length, &alloc);
		alloc = 1;
	}

	while (cur_length < length)
	{
		/* Get new buffer */
		buflen = 0;
		if ((lq = ldp_session_get_buffer(ls, &buflen)) == NULL)
		{
			LOG_SS_ERR("Addrlist message allocation failed in session %I:%d!", ls->ln->lsr_id, ls->space_id);
			return LDP_ERR;
		}

		buflen -= LDP_MESSAGE_HEADER_SIZE ;
		length = ldp_tlv_put_addrlist_initial(LDP_QDATA(lq), buflen, af, aptr + cur_length, length - cur_length, &bufused);
		/* bufused is size of used array data */
		cur_length += bufused;

		ldp_message_put_header(LDP_QHEADER(lq), ls, LDP_BIT_NONE, LDP_MSG_ADDR, length, NULL);
		ldp_session_send_message(ls, lq, LM_NONE);
	}

	return LDP_OK;
}

/**
 * ldp_session_make_lmap - fills buffer with complete LMAP/WITHDRAW message
 * @data: pointer to given buffer
 * @datasize: size of given buffer
 * @prefix: address to fill
 * @plen: prefix length
 * @label: locally-allocated label
 * @withdraw: indicate if this is label withdraw
 *
 * Writes to given buffer: message header, FEC TLV with single prefix,
 * Generic label TLV.
 * Buffer MUST have more than 8 + 12 + 8 = 28 bytes free (for IPv4)
 * or 8 + 24 + 8 = 40 bytes for IPv6 in worst case
 *
 * Returns number of bytes written.
 */
int
ldp_session_make_lmap(byte *data, size_t datasize, ldp_session *ls, ip_addr prefix, int plen, int label, int withdraw)
{
	int length;
	byte *x;

	x = data + LDP_MESSAGE_HEADER_SIZE;
	datasize -= LDP_MESSAGE_HEADER_SIZE;

	/* Fill in FEC TLV */
	length = ldp_tlv_fec_put_prefix(x + LDP_TLV_FEC_SIZE, datasize - LDP_TLV_FEC_SIZE, prefix, plen);
	length = ldp_tlv_put_fec(x, length);

	/* Fill in label TLV */
	length += ldp_tlv_put_generic_label(x + length, datasize - length, label);

	/* Fill in message header */
	length = ldp_message_put_header(data, ls, LDP_BIT_NONE, withdraw ? LDP_MSG_LWITH : LDP_MSG_LMAP, length, NULL);

	LLOG_SS_VRB("%s %I/%d label %d (total length %d)", withdraw ? "LWITHDRAW" : "LMAP", prefix, plen, label, length);

	return length;
}

/**
 * ldp_session_send_lmap_initial - Send all label mappings
 * @ls: pointer to LDP session
 *
 * Sends directly connected networks first witn IPv4NULL label
 * Sends other networks (stored in LFIB) of type transit.
 *
 */
static int
ldp_session_send_lmap_initial(ldp_session *ls)
{
	struct proto_ldp *pl;
	ldp_rib_group *rg;
	ldp_rib_entry *re;
	ldp_qmessage *lq = NULL;
	ldp_lib_database *lib;
	ldp_lib_transit *lt;
	int count, buflen = 0, bufused = 0;
	byte *buf = NULL;

	pl = ls->ln->pl;

	/* Announce directly connected networks at first */
	FIB_WALK(&pl->rib, rtmp)
	{
		rg = (ldp_rib_group *)rtmp;
		if ((re = (ldp_rib_entry *)HEAD(rg->entries)) == NULL)
			continue;

		if ((re->flags & LDP_RE_BEST) == 0)
			continue;

		/* Skip all non directly-connected */
		if (re->ls != NULL)
			continue;

		LLOG_DB_VRB("FIB PREFIX: %I/%d", rg->n.prefix, rg->n.pxlen);

		if (bufused + 100 >= buflen)
		{
			/* Time to send message */
			if (lq != NULL)
			{
				lq->length = bufused;
				ldp_session_send_message(ls, lq, LM_BULK);
			}

			buflen = bufused = 0;
			if ((lq = ldp_session_get_buffer(ls, &buflen)) == NULL)
			{
				LLOG_SS_VRB("Error allocating LMAP message for session %R:%d", ls->ln->lsr_id, ls->space_id);
				return LDP_ERR;
			}
			buf = LDP_QHEADER(lq);
		}

		bufused += ldp_session_make_lmap(buf + bufused, buflen - bufused, ls, rg->n.prefix, rg->n.pxlen, MPLS_LABEL_IPV4NULL, 0);
		lq->messages++;
	}
	FIB_WALK_END;

	/* Announce allocated labels for MPLS-enabled prefixes */
	lib = pl->lib;
	count = lib->size;
	FIB_WALK(&lib->f, ltmp)
	{
		lt = (ldp_lib_transit *)ltmp;

		if (lt->type != LDP_LIB_TYPE_TRANSIT)
			continue;

		LLOG_DB_VRB("Announcing %I/%d with label %d\n", lt->prefix, lt->pxlen, lt->me.local_label);

		if (bufused + 100 >= buflen)
		{
			/* Time to send message */
			if (lq != NULL)
			{
				lq->length = bufused;
				ldp_session_send_message(ls, lq, LM_BULK);
			}

			buflen = bufused = 0;
			if ((lq = ldp_session_get_buffer(ls, &buflen)) == NULL)
			{
				LLOG_SS_ERR("Error allocating LMAP message for session %I:%d", ls->ln->lsr_id, ls->space_id);
				return LDP_ERR;
			}
			buf = LDP_QHEADER(lq);
		}
		
		bufused += ldp_session_make_lmap(buf + bufused, buflen - bufused, ls, lt->prefix, lt->pxlen, lt->me.local_label, 0);
		lq->messages++;
	}
	FIB_WALK_END;

	/* Send remainings */
	LLOG_SS_VRB("TOTAL BUFUSED=%d", bufused);
	if (bufused > 0)
	{
		lq->length = bufused;
		ldp_session_send_message(ls, lq, LM_BULK);
	}

	return LDP_OK;
}

/*
 * ldp_dispatch_session_common - Common session dispatching code
 *
 */
static int
ldp_dispatch_session_common(ldp_session *ls, ldp_message_header *lmh)
{
	if ((ls->state != LDP_SSM_OPERATIONAL))
	{
		LLOG_SS_VRB("%s message received in session with wrong state, dropping session %R:%d %s", 
			ldp_message_texttype(lmh->type), ls->ln->lsr_id, ls->space_id, ldp_session_textstatus(ls->state));

		return LDP_STATUS_SESS_REJ_PADV; /* Not sure if this is right error code */
	}

	ls->last_received = CurrentTime;


	return LDP_STATUS_SUCCESS;
}

/**
 * ldp_dispatch_hello - dispatch hello message
 * @ln: ldp neighbour
 * @lh: ldp hello
 * @lmh: ldp message header
 * @tlv list: list of TLVs in this message
 *
 * Creates new LDP session for given neigbour if not already exists
 * Updates timers
 *
 * Returns LDP_OK or LDP_ERR in case of error
 */
int
ldp_dispatch_hello(ldp_neighbour *ln, ldp_hello *lh, ldp_message_header *lmh, list *tlv_list)
{
	void *data;
	int length;
	sock *sk;
	ip_addr remote;
	ldp_tlv_ipv4transport *tlv4;
	ldp_session *ls;
	struct ldp_iface *li = lh->li;
	struct proto_ldp *pl = ln->pl;
	struct ldp_config *lc = PL_TO_CF(pl);

	LLOG_HL_DBG("Got Hello from %I (LSR %R:%d) %p", lh->remote, ln->lsr_id, lh->space_id, ln);
	lh->last_packet = CurrentTime;

	/* Re-start hold timer */
	tm_stop(lh->hold_timer);
	tm_start(lh->hold_timer, lc->dhold_interval);

	/* Prepare remote router address */

	/* XXX: IPv6 ! */
	if ((tlv4 = ldp_tlv_find(tlv_list, LDP_TLV_IP4TRADDRESS, ldp_tlv_ipv4transport)) != NULL)
	{
		LLOG_HL_DBG("Found IP4 TRANSPORT TLV, sending reply to host %I", tlv4->addr);
		remote = tlv4->addr;
	}
	else
	{
		LLOG_HL_DBG("sending reply to host %I",  lh->remote);
		remote = lh->remote;
	}

	/* Session already exists, returning. XXX: We should search it by lsr_id & space_id */
	if (ldp_session_find(pl, ln->lsr_id, lh->space_id) != NULL)
		return LDP_OK;

	sk = li->tx_sk;

	if (sk->tbuf != sk->tpos)
		LLOG_HL_DBG("Old data overwritten on transmit");

	data = sk->tbuf;

	if ((length = ldp_hello_build(data, LDP_MAXPDU_INITIAL, pl)) <= 0)
	{
		LLOG_HL_VRB("ldp_build_hello() failed");
		return LDP_ERR;
	}

	/* Okay, let's send reply */
	sk_send_to(sk, length, LDP_ALL_ROUTERS, LDP_PORT);

	ls = ldp_session_add(lh, remote);
	ldp_session_prepare(ls);

	/* Set session labelspace */
	ls->space_id = lh->space_id;
	
	/* Check if we need to initiate TCP connect */
	if (ln->flags & LDP_FLAG_SESSION_ACTIVE)
	{
		LOG_HL_DBG("LDP session with %R:%d is in active state", ln->lsr_id, lh->space_id);
		ldp_session_init_connect(ls);
	}

	return LDP_OK;
}

int
ldp_dispatch_session_init(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list)
{
	struct proto_ldp *pl = ln->pl;
	struct ldp_config *lc = PL_TO_CF(pl);
	ldp_tlv_csess_params *tlv_sess;

	if (lmh->type != LDP_MSG_INIT)
	{
		LLOG_SS_DBG("Internal error, wrong function call for this message type");
		return LDP_STATUS_INT_ERROR;
	}

	if ((ls->state != LDP_SSM_OPENREC) && (ls->state != LDP_SSM_OPENSENT))
	{
		LOG_SS_DBG("INIT message received in session %I:%d with state %s, ignoring", ln->lsr_id, ls->space_id, ldp_session_textstatus(ls->state));
		return LDP_STATUS_SUCCESS; /* Simply ignore this message */
	}

	if ((tlv_sess = ldp_tlv_find(tlv_list, LDP_TLV_CSESSPAR, ldp_tlv_csess_params)) == NULL)
	{
		LOG_SS_DBG("No common session parameters in session %R:%d/INIT, dropping", ln->lsr_id, ls->space_id);
		return LDP_STATUS_MISS_MSG_PARAMS;
	}

	if (tlv_sess->pversion != LDP_PROTOCOL_VERSION)
	{
		LLOG_SS_DBG("Protocol version %d not supported in session %I:%d", tlv_sess->pversion, ln->lsr_id, ls->space_id);
		return LDP_STATUS_BAD_PROTO_VER;
	}

	/* Check LSR and labelspace */
	if ((tlv_sess->lsr_id != pl->router_id) || (tlv_sess->space_id != ls->space_id))
	{
		LOG_SS_DBG("LSR id differs between session and session params TLV: (s) %R:%d (t) %R:%d", pl->router_id, ls->space_id, tlv_sess->lsr_id, tlv_sess->space_id);
		return LDP_STATUS_BAD_LDP_ID;
	}

	/* Determine common PDU */
	ls->max_pdu = MIN(ln->datasize, (tlv_sess->max_pdu > 255 ? tlv_sess->max_pdu : 4096));
	LOG_SS_DBG("Common max PDU is %d", ls->max_pdu);

	/* Determine common keepalive */
	ls->keep_interval = MIN(lc->skeep_interval, tlv_sess->keep_interval);

	/* Send session INIT if we're in PASSIVE state */
	if (ls->state == LDP_SSM_OPENREC)
		ldp_session_send_init(ls);

	/* In any case, send keepalive indicating sessions parameters are accepted */
	ldp_session_send_keepalive(ls);
	/* Send our current address list */
	ldp_session_send_addrlist_initial(ls);
	/* Send current LMAPs */
	ldp_session_send_lmap_initial(ls);

	/* Consider session as operational */
	ldp_session_set_state(ls, LDP_SSM_OPERATIONAL);

	return LDP_STATUS_SUCCESS;
}

int
ldp_dispatch_session_keepalive(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list)
{
	int res;

	if ((res = ldp_dispatch_session_common(ls, lmh)) != LDP_STATUS_SUCCESS)
		return res;

	/* Do nothing since timer restart is done in ldp_dispatch_session_common() */

	return LDP_STATUS_SUCCESS;
}

int
ldp_dispatch_session_addrlist(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list)
{
	struct proto_ldp *pl = ln->pl;
	ldp_tlv_addrlist *tlv;
	ip_addr *addr;
	ldp_addrmap *la;
	int res;
	int af_size, count, i;

	if ((res = ldp_dispatch_session_common(ls, lmh)) != LDP_STATUS_SUCCESS)
		return res;

	if ((tlv = ldp_tlv_find(tlv_list, LDP_TLV_ADDRLIST, ldp_tlv_addrlist)) == NULL)
	{
		LOG_SS_DBG("No %s tlv in message %s (session %R:%d)", 
				ldp_tlv_texttype(LDP_TLV_ADDRLIST), ldp_message_texttype(lmh->type), ln->lsr_id, ls->space_id);
		return LDP_STATUS_MISS_MSG_PARAMS;
	}

	/* Determine how much addresses do we have */
	af_size = (tlv->iana_af == AFN_IP6) ? 16 : 4;
	count = (tlv->_base.length - 2) / af_size;

	addr = (ip_addr *)((byte *)tlv + BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN));

	/* Announce address message */
	if (lmh->type == LDP_MSG_ADDR)
	{
		/* Add addresses to list */
		for (i = 0; i < count; i++, addr++)
		{
			/* Check if next hop address is not used */
			if ((la = ldp_addrmap_find(pl, *addr)) != NULL)
			{
				if (la->ls != ls)
				{
					LOG_SS_ERR("Next-hop address %I already advertised by LSR %R:%d, overriding!", *addr, la->ls->ln->lsr_id, la->ls->space_id);
					ldp_addrmap_del(pl, la->ls, *addr);
				}
				else
				{
					/* It is common to periodically send full inteface list in some LDP implementation */
					//LOG_SS_DBG("Next-hop address %I already advertised!", *addr);
					continue;
				}

			}

			/* Not found, adding new record */
			if (ldp_addrmap_add(pl, ls, *addr) == LDP_ERR)
			{
				LLOG_SS_ERR("Address list map allocation failed!");
				return LDP_STATUS_INT_ERROR;
			}
	
			LOG_SS_DBG("Adding next-hop address %I to %R:%d", *addr, ln->lsr_id, ls->space_id);
		}
	}
	else if (lmh->type == LDP_MSG_ADDRWITH)
	{
		/* Delete addresses from list */
		for (i = 0; i < count; i++, addr++)
		{
			/* Check if next hop address is not used */
			if ((la = ldp_addrmap_find(pl, *addr)) == NULL)
			{
				/* This is not an error according to RFC 5036, 3.5.5.1 */
				continue;
			}

			/* Thsi can possibly happen if we get 2 different session advertising the same nexthop */
			if (la->ls != ls)
			{
				LOG_SS_ERR("Next-hop address %I advertised by LSR %R:%d, not withdrawing!", *addr, la->ls->ln->lsr_id, la->ls->space_id);
				continue;
			}

			LOG_SS_VRB("Withdrawing %I from %R:%d", *addr, ln->lsr_id, ls->space_id);
			ldp_addrmap_del(pl, ls, *addr);
		}
	}
	else
		return LDP_STATUS_INT_ERROR;
			
	return LDP_STATUS_SUCCESS;
}

int
ldp_dispatch_session_lmap(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list, list *forward_list)
{
	struct proto_ldp *pl = ln->pl;
	ldp_session *lnext;
	ldp_tlv_fec *fe;
	ldp_tlv_fec_element *fec;
	ldp_tlv_generic_label *gl;
	ldp_rib_entry *re;
	ldp_lib_general *lg;
	ldp_addrmap *la;

	ip_addr addr;
	int res, count, withdraw;
	u32 label = 0;
	net *n;

	if ((res = ldp_dispatch_session_common(ls, lmh)) != LDP_STATUS_SUCCESS)
		return res;

	/* Deternine what message we are dispatching */
	withdraw = (lmh->type == LDP_MSG_LWITH) ? 1 : 0;

	if ((fe = ldp_tlv_find(tlv_list, LDP_TLV_FEC, ldp_tlv_fec)) == NULL)
	{
		LOG_SS_DBG("No %s tlv in message %s (session %s:%d)", 
				ldp_tlv_texttype(LDP_TLV_FEC), ldp_message_texttype(lmh->type), ln->lsr_id, ls->space_id);
		return LDP_STATUS_MISS_MSG_PARAMS;
	}

	/* This TLV is optional for WITHDRAW message */
	if ((gl = ldp_tlv_find(tlv_list, LDP_TLV_GLABEL, ldp_tlv_generic_label)) == NULL)
	{
		if (withdraw == 0)
		{
			LOG_SS_DBG("No %s tlv in message %s (session %s:%d)!",
				ldp_tlv_texttype(LDP_TLV_GLABEL), ldp_message_texttype(lmh->type), ln->lsr_id, ls->space_id);
			return LDP_STATUS_MISS_MSG_PARAMS;
		}
	}
	else
		label = gl->label;

	/* XXX: label request part here */

	/* XXX: Do loop detection */

	count = fe->count;

	fec = (ldp_tlv_fec_element *)((byte *)(fe) + BIRD_ALIGN(sizeof(*fe), CPU_STRUCT_ALIGN));
	addr = fec->prefix;

	LOG_SS_DBG("Dispatching %d FEC prefixes", fe->count);

	/* Foreach prefix in FEC */
	while (count > 0)
	{
		/* XXX: IPv6, as usual */
		if (fec->type == LDP_FEC_ELEMENT_WILD)
		{
			/* XXX: do nothing */
			count--;
			continue;
		}

		/* At the moment we've got REMOTE ldp label and one of FEC prefixes mapped to that label */
		LOG_DB_DBG("Got prefix %I/%d with label %s via %R:%d", addr, fec->plen, ldp_label_texttype(label), ln->lsr_id, ls->space_id);

		if ((re = ldp_rib_search_entry(pl, addr, fec->plen, ls)) != NULL)
		{
			/* Already got some entry from peer, check if remote label differs (or this is WITHDRAW message) */
			if ((re->me.stack[0] != label) || (withdraw))
			{

				LOG_DB_DBG("Found old RIB entry %I/%d with label %s flags %X, removing", re->rg->n.prefix, re->rg->n.pxlen, 
						ldp_label_texttype(re->me.stack[0]), re->flags);

				/* 
				 * We have to announce label release to others if this was best route.
				 * Actually we can skip re->label. (but we are polite).
				 * re->label is either vali label from lib or ImpNULL (set in route.c)
				 */
				if (re->flags & LDP_RE_BEST)
					ldp_session_announce_lmap(pl, re, ls, LM_NONE, 1);

				/* Remove from kernel FIB */
				if (re->flags & LDP_RE_INSTALLED)
				{
					n = fib_find(&pl->proto.table->fib, &addr, fec->plen);
					if (n && n->routes)
						ldp_rt_update(pl, n, n->routes, re);
				}

				/* Remove from kernel LFIB */
				if (re->flags & LDP_RE_HASLABEL)
				{
					/* 
					 * Remove from forwarding use 
					 * XXX: Check if label is really installed.
					 */
					if ((lg = ldp_lib_label_find(pl, re->label)) != NULL)
						ldp_rt_withdraw_label(pl, &lg->me);
					ldp_lib_label_del(pl, re->label);
				}

				/* Delete prefix from RIB */
				ldp_rib_del_entry(pl, re); // re is freed here
				re = NULL;
			}
			else
			{
				/* Remote label is not changed. Let's move to next prefix */
				count--;
				fec++;
				addr = fec->prefix;
				continue;
			}
		}

		if (withdraw)
			continue;

		/* re is NULL here, so let's add new entry */
		re = ldp_rib_add_entry(pl, addr, fec->plen, ls);
		/* Update/set label for given re */
		re->me.action = KMPLS_OP_PUSH;
		re->me.local_label = 0;
		re->me.depth = 1;
		re->me.stack[0] = label;

		/* Try to match nextop from routing entry with one of LDP peers */
		lnext = NULL;
		if (((n = fib_find(&pl->proto.table->fib, &addr, fec->plen)) != NULL) && (n->routes != NULL))
		{
			/* Map nexthop of currently active route to neighbour */
			la = ldp_addrmap_find(pl, n->routes->attrs->gw);
			if (la != NULL)
				lnext = la->ls;
		}

		if (ls != lnext)
		{
			/* Nexthop is poiting to another session (or to non-LDP neighbour). Do nothing */
			LOG_SS_DBG("label %s on %I/%d: NOT ACTIVE", ldp_label_texttype(label), addr, fec->plen);
		}
		else
		{
			LOG_SS_DBG("label %s on %I/%d: INSTALLING", ldp_label_texttype(label), addr, fec->plen);
			/* 
			 * General handling code for new prefix+label pair.
			 * New label allocation, re + lib linking, lmap announce is done here
			 */
			ldp_session_install_prefix(pl, ls, re, n, n->routes, LDP_FLAG_INSTALL_RIB | LDP_FLAG_INSTALL_LFIB);
		}

		count--;
		fec++;
		addr = fec->prefix;
	}

	return LDP_STATUS_SUCCESS;
}

int
ldp_dispatch_session_lreq(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list, list *forward_list)
{
	ldp_tlv_fec *fe;
	int res;

	/* 
	 * Since we operate in Unsolicited mode and our peers are required to 
	 * operate in Unsolicited mode too we can't answer immediately only if
	 * we havent received label for newly-arrived IGP prefix
	 */

	if ((res = ldp_dispatch_session_common(ls, lmh)) != LDP_STATUS_SUCCESS)
		return res;

	if ((fe = ldp_tlv_find(tlv_list, LDP_TLV_FEC, ldp_tlv_fec)) == NULL)
	{
		LOG_SS_DBG("No %s tlv in message %s (session %R:%d)",
				ldp_tlv_texttype(LDP_TLV_FEC), ldp_message_texttype(lmh->type), ln->lsr_id, ls->space_id);
		return LDP_STATUS_MISS_MSG_PARAMS;
	}

	/* Check if prefix exists in our routing table */

	// ...

	return LDP_STATUS_SUCCESS;
}

int
ldp_dispatch_session_lrel(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list, list *forward_list)
{
	ldp_tlv_fec *fe;
	int res;

	if ((res = ldp_dispatch_session_common(ls, lmh)) != LDP_STATUS_SUCCESS)
		return res;

	if ((fe = ldp_tlv_find(tlv_list, LDP_TLV_FEC, ldp_tlv_fec)) == NULL)
	{
		LOG_SS_DBG("No %s tlv in message %s (session %R:%d)",
				ldp_tlv_texttype(LDP_TLV_FEC), ldp_message_texttype(lmh->type), ln->lsr_id, ls->space_id);
		return LDP_STATUS_MISS_MSG_PARAMS;
	}

	/* Let's simply ignore such request */

	return LDP_STATUS_SUCCESS;
}

int
ldp_dispatch_session_notify(ldp_neighbour *ln, ldp_session *ls, ldp_message_header *lmh, list *tlv_list, list *forward_list)
{
	ldp_tlv_status_params *lst;
	int res;

	if ((res = ldp_dispatch_session_common(ls, lmh)) != LDP_STATUS_SUCCESS)
		return res;

	if ((lst = ldp_tlv_find(tlv_list, LDP_TLV_STATUS, ldp_tlv_status_params)) == NULL)
	{
		LOG_SS_DBG("No %s tlv in message %s (session %R:%d)",
				ldp_tlv_texttype(LDP_TLV_STATUS), ldp_message_texttype(lmh->type), ln->lsr_id, ls->space_id);
		return LDP_STATUS_MISS_MSG_PARAMS;
	}

	if (lst->fatal)
	{
		LOG_SS_ERR("Fatal notification received in session with %R:%d, closing", ln->lsr_id, ls->space_id);
		ldp_session_shutdown(ls);
		return LDP_STATUS_SUCCESS;
	}

	/* Let's simply ignore such request */
	LOG_SS_DBG("NOTIFICATION message received in session with %R:%d, ignoring", ln->lsr_id, ls->space_id);

	return LDP_STATUS_SUCCESS;
}

