/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * BIRD routing table hooks / manipulation routines.
 *
 */

#define LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"
#include "ldp_proto.h" /* LDP_LABEL_IMPNULL */

static rte *ldp_kernel_add_sockaddr(ldp_rib_entry *re, rte *old, int readonly);
static rte *ldp_kernel_del_sockaddr(rte *old);

/*
 * Handles PRE-optimal route announce for main IGP route table.
 * It is ALWAYS called before RA_OPTIMAL hook
 */
void
ldp_rt_notify_pre(struct proto *p, rtable *tbl UNUSED, net *n, rte *new, rte *old, ea_list *attrs)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	int withdraw , dest_type;
	rte *r;
	ldp_addrmap *la;
	ldp_rib_entry *re;
	ldp_rib_group *rg;
	ldp_session *ls;

	DEBUG_PROTO(pl);

	/* Hook is called BEFORE new route gets installed */
	r = (new == NULL) ? old : new;
	withdraw = (new == NULL) ? 1 : 0;

	dest_type = r->attrs->dest;

	LDBG("-- PRE -- prefix %s %I/%d", withdraw ? "withdrawal" : "announce", n->n.prefix, n->n.pxlen);

	/* 
	 * Check if we need to delete old 'best' route from RIB/LFIB 
	 */
	if ((old != NULL) && ((rg = ldp_rib_search_group(pl, n->n.prefix, n->n.pxlen)) != NULL))
	{
		/* WE're interested in first entry only */
		WALK_LIST(re, rg->entries)
		{
			if (!(re->flags & LDP_RE_BEST))
				break;

			/* Remove flag & label if exists */
			re->flags &= ~LDP_RE_BEST;

			/* Notify all neighbours */
			ldp_session_announce_lmap(pl, re, NULL, LM_NONE, 1);

			if (re->flags & LDP_RE_HASLABEL)
			{
				/* XXX: Remove label from kernel */
				ldp_lib_label_del(pl, re->label);
				/* 
				 * Label from IP route will get removed automatically 
				 * because kernel protocol use RTM_DELETE | RTM_ADD
				 * instead of RTM_CHANGE
				 */
				re->flags &= ~(LDP_RE_HASLABEL | LDP_RE_INSTALLED);
				re->label = 0;
			}

			/* Remove entry if this is device route */
			if (re->ls == NULL)
				ldp_rib_del_entry(pl, re);
		}
	}

	if (withdraw == 1)
	{
		/* Old best route deleted, nothing left to do */
		UNDEBUG_PROTO();
		return;
	}

	switch (dest_type)
	{
		case RTD_ROUTER:
			/* 
			 * IGP or static route. 
			 * Check if nexthop record exists in LIB database 
			 */
			
			LLOG_DB_VRB("IGP %s %F gw %I", withdraw ? "withdrawal" : "announce", &n->n, r->attrs->gw);
			
			/* Map nexthop to neighbour */
			if ((la = ldp_addrmap_find(pl, r->attrs->gw)) == NULL)
			{
				/* 
				 * No ldp session found based on nexthop. Do nothing.
				 * XXX: We can add non-null label here to skip double lookup
				 */
				goto done;
			}
	
			ls = la->ls;
	
			LLOG_DB_DBG("Nexthop for %F mapped to LSR %R:%d", &n->n, ls->ln->lsr_id, ls->space_id);
	
			/* Check if prefix is advertised by LDP neighbour */
			if ((re = ldp_rib_search_entry(pl, n->n.prefix, n->n.pxlen, ls)) == NULL)
			{
				/* 
				 * FEC still not advertised in this session. We will add prefix after appropriate LMAP.
				 * Do nothing for now
				 */
				goto done;
			}
	
			/*  Prefix is already advertised in appropriate LDP session, let's prepare to install updated route */
	
			/* Check if stack saved in @re is acceptable */
			if ((re->me.action != KMPLS_OP_PUSH) || (re->me.depth == 0))
			{
				LLOG_DBG("Invalid label stack specification: action: %s, zero depth", ldp_mplsop_texttype(re->me.action));
				goto error;
			}
			
			/* 
			 * Install route to LFIB.
			 * Ask to send route to other neighbours.
			 * RIB update (with additional attribute) will be done by this hook
			 */
			ldp_session_install_prefix(pl, ls, re, n, new, LDP_FLAG_INSTALL_LFIB|LDP_FLAG_ANNOUNCE);
			/* Add LDP label to announced prefix */
			ldp_kernel_add_sockaddr(re, new, 0);
	
			LLOG_DB_DBG("Assigning LDP label %s to %F", ldp_label_texttype(re->label), &n->n);
			break;

		case RTD_MULTIPATH:
			/* XXX: Do something here */
			LLOG_DB_DBG("Ignoring multipath route %F", &n->n);
			break;

		case RTD_DEVICE:
		case RTD_BLACKHOLE:
			/* Route to interface */
			LLOG_DB_DBG("Direct route announce %F on %s", &n->n, r->attrs->iface->name);
			/* 
			 * Add route to RIB with IPv4 implicit NULL label.
			 * Assume no multipach for directly connected networks
			 * at the moment
			 */
			if ((re = ldp_rib_add_entry(pl, n->n.prefix, n->n.pxlen, NULL)) == NULL)
			{
				LOG_DB_VRB("Error allocating RIB entry for directly connected prefix %F", &n->n);
				goto error;
			}
	
			ldp_rib_set_best(pl, re);
			/* Announce to all sessions with implicit NULL */
			re->label = LDP_LABEL_IMPNULL;
			ldp_session_announce_lmap(pl, re, NULL, LM_NONE, 0);
			break;

#if 0
		case RTD_PROHIBIT:
		case RTD_UNREACHABLE:
			/* Do nothing */
			break;
		default:
			/* Do nothing*/
#endif
	}


done:
error:
	UNDEBUG_PROTO();
}

void
ldp_rt_announce_label(struct proto_ldp *pl, ldp_lib_general *lg, ip_addr nexthop, ldp_session *ls)
{
	rta *a, A;
	struct neighbor *nei;
	net *n;
	bird_mpls_entry lme;
	u32 label;
	char buf[32];
	eattr *ea;
	struct adata *ad;
	ea_list *tmpa;
	rte *e;
	
	addr_print_me(buf, sizeof(buf), &lg->me, 0);
	label = lg->me.local_label;
	if ((label < LDP_LABEL_MIN) || (label > LDP_LABEL_MAX))
	{
		LOG_DB_ERR("Label %s insert request is invalid. Ignoring", buf);
		return;
	}

	if ((nei = neigh_find(&pl->proto, &nexthop, NEF_STICKY)) == NULL)
	{
		LOG_DB_VRB("Label %s installation failed: nexthop %I is not directly reachable", buf, nexthop);
		return;
	}

	/* Prepare attribute */
	bzero(&A, sizeof(A));
	A.proto = &pl->proto;
	A.source = RTS_LDP;
	A.scope = SCOPE_UNIVERSE;
	A.cast = RTC_UNICAST;
	A.dest = RTD_ROUTER;
	A.flags = 0;
	A.gw = nexthop;
	A.iface = nei->iface;
	/* Prepare label extended attribute */
	
	/* Allocate temporary attribute for LA_MPLSENTRY */	
	tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
	memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
	tmpa->count = 1;

	/* Setup external storage for sockaddr  */
	ad = alloca(sizeof(struct adata) + sizeof(struct bird_mpls_entry));
	ad->length = sizeof(struct bird_mpls_entry);
	memcpy(ad + 1, &lg->me, sizeof(struct bird_mpls_entry));

	/* Finish attribute setup */
	ea = (eattr *)(tmpa + 1);
	ea->id = EA_CODE(EAP_LDP, LA_MPLSENTRY);
	ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;
	ea->u.ptr = ad;

	A.eattrs = tmpa;

	a = rta_lookup(&A);

	/* Get new prefix */
	memset(&lme, 0, sizeof(lme));
	lme.local_label = label;
	n = fib_get(&pl->mpls_table->fib, &lme, 0);

	e = rte_get_temp(a);
	e->net = n;
	e->pflags = 0;
	/* Setup protocol-specific attributes */
	if (ls != NULL)
		e->u.ldp.router_id = ls->ln->lsr_id;
	e->u.ldp.ltype = lg->type;

	LOG_DB_DBG("Installing label %s with nexthop %I", buf, nexthop);

	/* Do announce */
	rte_update(pl->mpls_table, n, &pl->proto, &pl->proto, e);
}

void
ldp_rt_withdraw_label(struct proto_ldp *pl, bird_mpls_entry *me)
{
	net *n;
	char buf[32];
	bird_mpls_entry lme;
	u32 label = me->local_label;

	addr_print_me(buf, sizeof(buf), me, 0);

	if ((label < LDP_LABEL_MIN) || (label > LDP_LABEL_MAX))
	{
		LOG_DB_ERR("Label %s withdraw request is invalid. Ignoring", buf);
		return;
	}

	memset(&lme, 0, sizeof(lme));
	lme.local_label = me->local_label;
	if ((n = fib_find(&pl->mpls_table->fib, &lme, 0)) == NULL)
	{
		LOG_DB_VRB("Label %d does not exist in LFIB, unable to withdraw", label);
		return;
	}

	/* Do withdraw*/
	rte_update(pl->mpls_table, n, &pl->proto, &pl->proto, NULL);
}

/**
 * ldp_rt_update - do IP routing table update adding/removing LA_MPLSENTRY attribute
 * @pl: pointer to protocol instance
 * @n: pointer to network we're updating
 * @old: pointer to current rte we need to update
 * @re: rib entry. NULL if we need to withdraw attribute
 *
 */
void
ldp_rt_update(struct proto_ldp *pl, net *n, rte *old, ldp_rib_entry *re)
{
	rte *new;
	char buf[32];

	if (re)
	{
		addr_print_me(buf, sizeof(buf), &re->me, 0);
		LOG_DB_DBG("Adding LDP label %s to prefx %F", buf, &n->n);
		new = ldp_kernel_add_sockaddr(re, old, 1);
	}
	else
	{
		LOG_DB_DBG("Removing LDP label from prefx %F", &n->n);
		new = ldp_kernel_del_sockaddr(old);
	}

	rte_update(pl->proto.table, n, new->attrs->proto, new->attrs->proto, new);
}

/**
 * ldp_kernel_add_sockaddr - Adds MPLS sockaddr to given rte
 * @re: pointer to RIB entry
 * @old: pointer to route entry we need to update
 * @readonly: signal if rte is readonly
 *
 * We can be called from2 different places:
 * 1) from PRE-announce hook. Here we can
 * a) alter rte the way we like since it is temporary
 * b) free old [cached] rta since it will not be used
 *
 * 2) from any place via rte_update() and current rte/rta residing
 * in appropriate fib. This is 'readonly' case
 * so we
 * a) obtain temporary rte
 * b) do more or less efficient rta copy with new attribute
 * c) do not free original rta
 *
 * We return pointer to new rte with new cached rta.
 */
static rte *
ldp_kernel_add_sockaddr(ldp_rib_entry *re, rte *old, int readonly)
{
	rta *a;
	eattr *ea;
	struct adata *ad;
	ea_list *tmpa;
	rte *new = old;

	a = old->attrs;

	/* Operate on rte copy */
	if (readonly)
	{
	  new = rte_get_temp(a);
	  memcpy(new, old, sizeof(rte));
	}

	/* Allocate temporary attribute for LA_MPLSENTRY */	
	tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
	memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
	tmpa->count = 1;

	/* Setup external storage for sockaddr  */
	ad = alloca(sizeof(struct adata) + sizeof(struct bird_mpls_entry));
	ad->length = sizeof(struct bird_mpls_entry);
	memcpy(ad + 1, &re->me, sizeof(struct bird_mpls_entry));

	/* Finish attribute setup */
	ea = (eattr *)(tmpa + 1);
	ea->id = EA_CODE(EAP_LDP, LA_MPLSENTRY);
	ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;
	ea->u.ptr = ad;

	/* Make shiny new attributes */
	new->attrs = rta_lookup_self(rta_copy_with_ea(a, tmpa));
	/* Free old rta if it is cached and we're called from PRE hook */
	if ((!readonly) && (a->aflags & RTAF_CACHED))
		rta_free(a);
	
	return new;
}

/**
 * ldp_kernel_del_sockaddr - Remove LA_MPLSENTRY attribute from rte copy
 * @_new: pointer to original rte
 *
 * Returnes pointer to rte which is full copy of supplied _rte with
 * LA_MPLSENTRY attribute stripped. rte contains cached rta
 */
static rte *
ldp_kernel_del_sockaddr(rte *old)
{
	rte *new;
	rta *a;
	ea_list *el, tmpl;
	eattr *ea;
	int i;

	if (!old->attrs->eattrs)
		return old;

	new = rte_get_temp(old->attrs);
	memcpy(new, old, sizeof(rte));
	/* Copy attributes from rte */
	/* Prepare empty attribute list */
	memset(&tmpl, 0, sizeof(tmpl));
	a = rta_copy_with_ea(new->attrs, &tmpl);
	el = a->eattrs;
	/* Check attributes */
	for (i = 0; i < el->count; i++)
	{
		ea = &el->attrs[i];
		if (ea->id != EA_CODE(EAP_LDP, LA_MPLSENTRY))
			continue;
		/* 
		 * Remove this code entirely. rta_copy allocates external storage
		 * from rta_pool memory block so we are freeing it
		 */
		mb_free(ea->u.ptr);
		el->count--;
		memmove(&el->attrs[i], &el->attrs[el->count], (el->count - i) * sizeof(eattr));
		break;
	}

	/* make cached rta */
	new->attrs = rta_lookup_self(a);

	return new;
}

#if 0
static void
ldp_rt_notify(struct proto *p, rtable *tbl UNUSED, net *n, rte *new, rte *old, ea_list *attrs)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;

	/* Do nithing since all we have to do is done in ldp_rtattr_notify currently */
	int withdraw = (new == NULL) ? 1 : 0;
	LP_LVERBOSE(LD_DB, "-- NOTIFY -- prefix %s %I/%d", withdraw ? "withdrawal" : "announce", n->n.prefix, n->n.pxlen);
}
#endif
