/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * LDP sessions state machines.
 * Transport session setup, message sending functions
 *
 * Packet transmitter operates the following way:
 * every packet is queued to neighbour internal sending queue (however immediate sending can be requested, too)
 * On timer (or some other event) periodic function tries to send this queue grouping messages
 * to PDUs and sending them via sk_send. IF bird can't immediately transmit requested buffer
 * (sk_send returns 0) LDP_FLAG_TRANSMIT_RUNNING flags is set stopping periodic function from running.
 * Flags is cleared on either _err or _tx socket hook. In latter case transmitting function is immediately
 * called again.
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"
#include "ldp_proto.h"

/*
 *              Session Initialization State Transition Diagram
 *
 *                              +------------+
 *                              |            |
 *                +------------>|NON EXISTENT|<--------------------+
 *                |             |            |                     |
 *                |             +------------+                     |
 *                | Session        |    ^                          |
 *                |   connection   |    |                          |
 *                |   established  |    | Rx any LDP msg except    |
 *                |                V    |   Init msg or Timeout    |
 *                |            +-----------+                       |
 *   Rx Any other |            |           |                       |
 *      msg or    |            |INITIALIZED|                       |
 *      Timeout / |        +---|           |-+                     |
 *   Tx NAK msg   |        |   +-----------+ |                     |
 *                |        | (Passive Role)  | (Active Role)       |
 *                |        | Rx Acceptable   | Tx Init msg         |
 *                |        |    Init msg /   |                     |
 *                |        | Tx Init msg     |                     |
 *                |        |    Tx KeepAlive |                     |
 *                |        V    msg          V                     |
 *                |   +-------+        +--------+                  |
 *                |   |       |        |        |                  |
 *                +---|OPENREC|        |OPENSENT|----------------->|
 *                +---|       |        |        | Rx Any other msg |
 *                |   +-------+        +--------+    or Timeout    |
 *   Rx KeepAlive |        ^                |     Tx NAK msg       |
 *      msg       |        |                |                      |
 *                |        |                | Rx Acceptable        |
 *                |        |                |    Init msg /        |
 *                |        +----------------+ Tx KeepAlive msg     |
 *                |                                                |
 *                |      +-----------+                             |
 *                +----->|           |                             |
 *                       |OPERATIONAL|                             |
 *                       |           |---------------------------->+
 *                       +-----------+     Rx Shutdown msg
 *                All other  |   ^            or Timeout /
 *                  LDP msgs |   |         Tx Shutdown msg
 *                           |   |
 *                           +---+
 */
static int ldp_session_setup_listen_socket(void);
static int ldp_session_send_buffer(ldp_session *ls, sock *sk, int length);
static void ldp_session_check_tranfers(timer *t);
static void ldp_session_connect_hook(sock *sk);
static int ldp_session_rx_hook(sock *sk, int size);
static void ldp_session_tx_hook(sock *sk);
static void ldp_session_send_keepalive_hook(timer *t);
static void ldp_session_hold_timeout_hook(timer *t);
static void ldp_session_err_hook(sock *sk, int err);
static int ldp_tcp_incoming_hook(sock *sk, int dummy UNUSED);
static void ldp_tcp_incoming_err_hook(sock *sk, int err);
static void ldp_session_shutdown_transport(ldp_neighbour *ln);
static enum LDP_SESS_ROLE ldp_session_get_role(struct ldp_iface *li, ip_addr remote);
static int ldp_session_hash(uint32_t addr, uint16_t space_id);
static int ldp_session_preprocess(ldp_neighbour *ln, byte *data, int size);
static int ldp_neighbour_hash(uint32_t addr);

/* Global listening TCP socket */
sock *tcp_sk;

/*
 * Set session FSM state
 */
enum LDP_SSM
ldp_session_set_state(ldp_session *ls, enum LDP_SSM state)
{
	LOG_SS_VRB("Session %R:%d state changed to %s", ls->lsr_id, ls->space_id, ldp_session_textstatus(state));
	if (ls->state == state)
		return state;

	ls->state = state;
	ls->last_state = CurrentTime;
	
	return state;
}

/*
 * Return text status for station state
 */
char *
ldp_session_textstatus(enum LDP_SSM state)
{
	char *buf = "";

	switch (state)
	{
		case LDP_SSM_NONEXISTENT:
			buf = "Non existent";
			break;
		case LDP_SSM_INITIALIZED:
			buf = "Init";
			break;
		case LDP_SSM_OPENSENT:
			buf = "OpenSent";
			break;
		case LDP_SSM_OPENREC:
			buf = "OpenRec";
			break;
		case LDP_SSM_OPERATIONAL:
			buf = "Operational";
			break;
		default:
			buf = "Unknown";
	}

	return buf;
}

/*
 * Open listening socket if this is not done alread.
 * Set session-related protocol option, start 
 * rtansmitter timer
 */
int
ldp_sessions_init(struct proto_ldp *pl)
{
	timer *t;
	int i;
	list *l;

	/* Allocate and init hashes for neighbours/sessions */
	if ((pl->neighbour_hash = mb_alloc(MAIN_POOL(pl), sizeof(list) * LDP_HASH_SIZE)) == NULL)
		return LDP_ERR;

	if ((pl->session_hash = mb_alloc(MAIN_POOL(pl), sizeof(list) * LDP_HASH_SIZE)) == NULL)
		return LDP_ERR;

	for (i = 0, l = pl->neighbour_hash; i < LDP_HASH_SIZE; i++, l++)
		init_list(l);

	for (i = 0, l = pl->session_hash; i < LDP_HASH_SIZE; i++, l++)
		init_list(l);

	init_list(&pl->neighbour_list);
	init_list(&pl->session_list);

	if (tcp_sk == NULL)
	{
		if (ldp_session_setup_listen_socket() != LDP_OK)
			return LDP_ERR;
	}


	/* Set IP address that we will advertise in IPv4 transport TLV */
	pl->tcp_saddr = ldp_session_global_saddr(pl);

	/* Setup periodic trasmit timer */
	t = tm_new(MAIN_POOL(pl));
	t->hook = ldp_session_check_tranfers;
	t->data = pl;
	t->recurrent = 1;
	pl->xmit_timer = t;
	tm_start(t, 1);


	return LDP_OK;
}

/*
 * Called on protocol shutting down. Free socket if this was the last
 * instance
 */
void
ldp_sessions_shutdown(struct proto_ldp *pl, int last)
{
	ldp_session *ls, *ls_next;
	node *n;

	rfree(pl->xmit_timer);
	pl->xmit_timer = NULL;

	/* Ensure no sessions are left */
	WALK_LIST_DELSAFE(ls, ls_next, pl->session_list)
	{
		ls = SKIP_BACK(ldp_session, n_list, ls);
		ldp_session_shutdown(ls);
	}

	mb_free(pl->neighbour_hash);
	mb_free(pl->session_hash);

	pl->neighbour_hash = pl->session_hash = NULL;

	/* Check lists */
	WALK_LIST(n, pl->neighbour_list)
		LOG_ERR("Neighbour %p[+-] remains!", n);
	WALK_LIST(n, pl->session_list)
		LOG_ERR("Session %p[+-] remains!", n);

	if (last)
	{
		rfree(tcp_sk);
		tcp_sk = NULL;
	}
}

/**
 * ldp_session_notify_reconfigure - notify LDP neighbours about configuration change
 *
 * @pl: LDP protocol instance
 *
 * Dispatches notification about our configuration change
 * FIXME: We should send a note to all neighbours
 */
void
ldp_session_notify_reconfigure(struct proto_ldp *pl)
{
	LOG_SS_VRB("Protocol reconfigure called, let's notify active sessions");

	pl->cfg_index++;
}

/**
 * ldp_session_global_saddr - get IP address for transport connections
 * @pl: pointer to protocol instance
 *
 * Returns either global router ID (if such an address resides on some interface)
 * or first address of first interface in list. IPA_NONE is returned if no interfaces
 * are available
 * XXX: IPv6 case not handled
 */
ip_addr
ldp_session_global_saddr(struct proto_ldp *pl)
{
	ip_addr addr; 

	/* Check if this addr is used on some interface */
	struct iface *iface;
	struct ifa *a, *f = NULL;
	
	addr = ipa_to_u32(config->router_id);

	WALK_LIST(iface, iface_list)
	{
		if (!(iface->flags & IF_UP))
			continue;

		WALK_LIST(a, iface->addrs)
		{
			if (f = NULL)
				f = a;
			if (ipa_compare(addr, a->ip) == 0)
			{
				LOG_DBG("Found router-id %I on interface %s", addr, iface->name);
				return addr;
			}
		}
	}

	if (f == NULL)
	{
		LOG_ERR("No interface adresses found! Are we running on a toaster?!");
		return IPA_NONE;
	}

	/*
	 * RFC 5036, clause 2.5.2
	 *  An LSR MUST advertise the same transport address in all Hellos that
	 *  advertise the same label space.  This requirement ensures that two
	 *  LSRs linked by multiple Hello adjacencies using the same label spaces
	 *  play the same connection establishment role for each adjacency.
	 */

	LOG_ERR("router-id address %I not found on any interface, using first found adress: %I", addr, f->ip);
	return f->ip;
}

/**
 * ldp_session_setup - Create new session structure based on hello
 * @li: pointer to hello session
 * @remote: destination session address
 *
 * Creates new session filling up all needed pointers and determining session type
 * Session FSM starts from here
 *
 */
ldp_session *
ldp_session_add(ldp_hello *lh, ip_addr remote)
{
	ldp_session *ls;
	ldp_neighbour *ln = lh->ln;
	struct proto_ldp *pl = ln->pl;

	if ((ls = mb_allocz(MAIN_POOL(pl), sizeof(ldp_session))) == NULL)
	{
		LOG_SS_ERR("Session init with %R:%d failed: memory allocation failure", lh->ln->lsr_id, lh->space_id);
		return NULL;
	}

	ls->lsr_id = ln->lsr_id;
	ls->space_id = lh->space_id;
	ls->li = lh->li;
	ls->ln = lh->ln;

	/* Set max PDU to defaul PDU size */
	ls->max_pdu = LDP_MAXPDU_INITIAL;

	ls->state = LDP_SSM_NONEXISTENT;

	/* Add to hash */
	add_tail(&pl->session_hash[ldp_session_hash(ls->lsr_id, ls->space_id)], NODE ls);
	/* Add to global list */
	add_tail(&pl->session_list, &ls->n_list);
	/* Add to per-lsr list */
	add_tail(&ln->session_list, &ls->n_lsr);

	LDBG("Added session %p[%p] to neighbour %p", ls, &ls->n_list, ln);

	/* Set initial transport session parameters if needed */
	if (ln->tr_state == LDP_SESSION_TRANSPORT_NONE)
	{
		/* Setup local/remote addresses */
		ln->local = pl->tcp_saddr;
		ln->remote = remote;

		if (ldp_session_get_role(lh->li, remote) == LDP_SESSION_ACTIVE)
		{
			ln->flags |= LDP_FLAG_SESSION_ACTIVE; /* Flag is set on neighbour, not session */
			/* Connect is initiated in ldp_dispatch_hello() from where we are called */
		}
	}

	/* Check if transport connection is already UP */
	if (ln->tr_state == LDP_SESSION_TRANSPORT_ACTIVE)
	{
		if (ln->flags & LDP_FLAG_SESSION_ACTIVE)
		{
			/* Transmit INIT */
			ldp_session_set_state(ls, LDP_SSM_OPENSENT);
		}
		else
			ldp_session_set_state(ls, LDP_SSM_OPENREC);
	}

	return ls;
}

/**
 * ldp_session_shutdown - Close LDP session
 *
 * Removes logical session
 */
void
ldp_session_shutdown(ldp_session *ls)
{
	ldp_neighbour *ln = ls->ln;
	struct proto_ldp *pl = ln->pl;
	ldp_rib_entry *re;
	ldp_rib_group *rg;
	ldp_lib_transit *lt, *lt_next;
	net *n;

	if (_IS_DELETED(ls))
		return;
	_SET_DELETED(ls);

	LLOG_SS_VRB("Shutting down logical LDP session with %R:%d", ln->lsr_id, ls->space_id);

	/* 
	 * We're trying to send shutdown notification for exery session if this is possible.
	 * If socket internal queue is empty this means last PDU was sent completely and we can
	 * send PDUs with NOTIFICATION message for every session. If not, this possibly means our
	 * packets got queued (sk_send returns 0) and never get send due to rfree() immediately following.
	 * However this is much better than nothing
	 */

	if ((ln->tr_state == LDP_SESSION_TRANSPORT_ACTIVE) && (ln->sk->tbuf == ln->sk->tpos))
	{
		// FIXME: send notification
	}

	/* Kill timers */
	rfree(ls->hold_timer);
	rfree(ls->send_timer);

	/* Delete from hash */
	rem_node(NODE ls);
	/* Delete from global list */
	rem_node(&ls->n_list);
	/* Delete from neighbour list */
	rem_node(&ls->n_lsr);

	/* Free all nexthop addresses associated with session */
	ldp_addrmap_flush_session(pl, ls);

	/* Remove all MPLS-enabled prefixes associated with session */
	WALK_LIST_FIRST(re, ls->prefix_list)
	{
		re = SKIP_BACK(ldp_rib_entry, n_lss, re);
		rg = re->rg;
		if ((re->flags & LDP_RE_INSTALLED) && (n = fib_find(&pl->proto.table->fib, FPREFIX(&rg->n), rg->n.pxlen)) != NULL) {
			LOG_KR_DBG("Removing MPLS label from %F", &rg->n);

			/* Announcing new route without label */
			ldp_rt_update(pl, n, n->routes, NULL);
		}
		ldp_rib_del_entry(pl, re);
	}

	/* Free all labels associated with session */
	WALK_LIST_DELSAFE(lt, lt_next, ls->label_list)
	{
		lt = SKIP_BACK(ldp_lib_transit, n_lss, lt);
		LOG_DB_DBG("deleting label %s from session %R:%d", ldp_label_texttype(lt->me.local_label), ln->lsr_id, ls->space_id);
		ldp_lib_label_del(pl, lt->me.local_label);
	}

	/* Check if this was the last logical session with given neighbour */
	if (EMPTY_LIST(ln->session_list))
		ldp_session_shutdown_transport(ln);
	
	mb_free(ls);
}

/*
 * Hash labelspace to uint16_t
 */
static int 
ldp_session_hash(uint32_t lsr_id, uint16_t space_id)
{
	return ((ntohl(lsr_id) % 65536) | space_id);
}

/**
 * ldp_session_find - Finds session by router id and space id
 * @pl: pointer to LDP protocol instance
 * @lsr_id: router id
 * @space_id: labelspace id
 *
 * Returns pointer to LDP session of NULL
 */
ldp_session *
ldp_session_find(struct proto_ldp *pl, uint32_t lsr_id, uint16_t space_id)
{
	ldp_session *ls, *ls_next;

	WALK_LIST_DELSAFE(ls, ls_next, pl->session_hash[ldp_session_hash(lsr_id, space_id)])
	{
		if ((ls->lsr_id == lsr_id) && (ls->space_id == space_id))
			return ls;
	}

	return NULL;
}

/**
 * ldp_neighbour_add - Create, fill and add to neighbour to list
 * @pl: pointer to protocol
 * @lsr_id: LSR id in host format
 *
 * Creates new structure, initializes common fields and adds structure
 * to global hast/list
 */
ldp_neighbour *
ldp_neighbour_add(struct proto_ldp *pl, uint32_t lsr_id)
{
	ldp_neighbour *ln;

	ln = mb_allocz(MAIN_POOL(pl), sizeof(ldp_neighbour) + LDP_SESSION_BASE_BUFFER);
	ln->datasize = LDP_SESSION_BASE_BUFFER;
	ln->data = (byte *)(ln + 1);
	ln->lsr_id = lsr_id;
	ln->pl = pl;
	init_list(&ln->hello_list);
	init_list(&ln->session_list);
	init_list(&ln->packets_list);

	/* Add to hash */
	add_tail(&pl->neighbour_hash[ldp_neighbour_hash(ln->lsr_id)], NODE ln);
	/* Add to global list */
	add_tail(&pl->neighbour_list, &ln->n_list);

	LOG_SS_VRB("New neighbour: %R:*", ln->lsr_id);
	LDBG("NEW NEIGHBOUR: %R:* %p", ln->lsr_id, ln);

	return ln;
}

/**
 * ldp_neighbour_shutdown - Close and free LDP neighbour
 * @ln: pointer to neighbour
 *
 * Removes all hello, logical and transport sessions and destroys neighbour.
 * Protected by recursive calls
 *
 */
void
ldp_neighbour_shutdown(ldp_neighbour *ln)
{
	ldp_hello *lh, *lh_next;

	if (_IS_DELETED(ln))
		return;
	_SET_DELETED(ln);

	LOG_SS_VRB("Shutting down neighbour %R:*", ln->lsr_id);

	/* Check if transport session is alive and close all logical sessions */
	ldp_session_shutdown_transport(ln);

	/* Close all hello adjacencies */
	WALK_LIST_DELSAFE(lh, lh_next, ln->hello_list)
	{
		lh = SKIP_BACK(ldp_hello, n_lsr, lh);
		LOG_HL_DBG("Preparing to delete hello adjacency %R:%d", ln->lsr_id, lh->space_id);
		ldp_hello_shutdown(lh);
	}
	
	/* Free transport data buffer if allocated separately */
	if (ln->flags & LDP_FLAG_CUSTOM_BUF)
		mb_free(ln->data);

	/* Delete from hash */
	rem_node(NODE ln);
	/* Delete from global list */
	rem_node(&ln->n_list);

	mb_free(ln);
}

static int 
ldp_neighbour_hash(uint32_t addr)
{
	return (ipa_hash(addr) % LDP_HASH_SIZE);
}

/**
 * ldp_neighbour_find - Finds neighbour by its router ID
 * @pl: pointer to LDP protocol instance
 * @addr: router id in host format
 *
 */
ldp_neighbour *
ldp_neighbour_find(struct proto_ldp *pl, uint32_t addr)
{
	ldp_neighbour *ln, *ln_next;

	WALK_LIST_DELSAFE(ln, ln_next, pl->neighbour_hash[ldp_neighbour_hash(addr)])
	{
		if (ln->lsr_id == addr)
			return ln;
	}

	return NULL;
}

/**
 * ldp_session_install_prefix - creates local label mapping for given prefix and send updated routes to kernel
 * @pl: pointer to LDP protocol
 * @ls: pointer to LDP session
 * @re: pointer to RIB entry
 * @n: pointer to prefix
 * @e: pointer to best RTE in this prefix (usually n->routes)
 * flags: announcement flags
 *
 * Creates local label mapping for given prefix, links @re with new mapping, adds label to session label list.
 * if LDP_FLAG_INSTALL_RIB is set, kernel route update for IP prefix is executed.
 * if LDP_FLAG_INSTALL_LFIB is set, label is added to AF_MPLS family.
 * Assumes all data already checked. @re must contain remote label in me.stack[0]
 *
 * Returns standart LDP message dispathing status
 *
 */
int
ldp_session_install_prefix(struct proto_ldp *pl, ldp_session *ls, ldp_rib_entry *re, net *n, rte *e, int flags)
{
	ldp_lib_transit *lt;
	u32 label;

	label = re->me.stack[0];

	if (label == MPLS_LABEL_IMPLNULL)
	{
		/* This label must not appear on the wire and stack popping is required */
		flags &= ~LDP_FLAG_INSTALL_RIB;
	}

	/* INSTALL route into kernel */
	if (flags & LDP_FLAG_INSTALL_RIB)
	{
		LOG_DB_DBG("Installing remote label %s on %I/%d", ldp_label_texttype(label), n->n.prefix, n->n.pxlen);
		/* Assume n->routes is not empty and valid */
		ldp_rt_update(pl, n, e, re);
		re->flags |= LDP_RE_INSTALLED;
	}

	/* Add entry to LFIB and propagade it to neighbours */

	/* Create and fill MPLS label for LFIB */
	if ((lt = (ldp_lib_transit *)ldp_lib_label_add(pl, LDP_LIB_TYPE_TRANSIT, KMPLS_OP_SWAP)) == NULL)
	{
		LOG_DB_ERR("Out of memory on label allocation to %I/%d", n->n.prefix, n->n.pxlen);
		return LDP_STATUS_INT_ERROR;
	}

	lt->ls = ls;
	lt->nexthop = e->attrs->gw;
	lt->prefix = n->n.prefix;
	lt->pxlen = n->n.pxlen;
	/* Add label to per-session list */
	add_tail(&ls->label_list, &lt->n_lss);

	/* Save newly-allocated LOCAL label to RIB entry */
	re->label = lt->me.local_label;
	/* Set appropriate flags for RIB. */
	re->flags |= LDP_RE_HASLABEL;
	ldp_rib_set_best(pl, re);

	if (flags & LDP_FLAG_INSTALL_LFIB)
	{
		LOG_DB_DBG("Assigning local LDP label %s to %I/%d", 
				ldp_label_texttype(re->label),
				n->n.prefix, n->n.pxlen);
		/* INSTALL label into kernel */
		ldp_rt_announce_label(pl, (ldp_lib_general *)lt, lt->nexthop, ls);
	}

	/* 
	 * Announce label to other neighbours.
	 */
	ldp_session_announce_lmap(pl, re, ls, LM_NONE, 0);

	return LDP_STATUS_SUCCESS;
}

/**
 * ldp_session_announce_lmap - Sends local label mapping/withdraw to all neighbours
 * @pl: pointer to LDP protocol
 * @re: pointer to RIB entry
 * @one: pointer to LDP session not to send LMAP to
 * @flags: flags to pass to ldp_session_sendto_all_butone
 * @withdraw: indicate if this is label withdraw
 *
 */
void
ldp_session_announce_lmap(struct proto_ldp *pl, ldp_rib_entry *re, ldp_session *one, int flags, int withdraw)
{
	int buflen = LDP_QHEADER_SIZE + LDP_MAXPDU_INITIAL;
	ldp_qmessage *lq = alloca(buflen);
	lq->messages = 1;

	lq->length = ldp_session_make_lmap(LDP_QHEADER(lq), LDP_MAXPDU_INITIAL, NULL, re->rg->n.prefix, re->rg->n.pxlen, re->label, withdraw);

	ldp_session_sendto_all_butone(pl, lq, one, flags|LM_NOFREE);
}

/**
 * ldp_session_get_buffer - Creates new message buffer suitable to feed to ldp_send_message
 * @ls: pointer to LDP session
 * @len: pointer to buffer size
 *
 * Creates new message buffer filling in space_id. Buffer linking is done in
 * ldp_send_message if nessessary. NULL pointer can be supplied as @len. In that case
 * @ls->max_pdu - LDP_PDU_HEADER_SIZE is allocated. This is guaranteed to be at least 256 bytes
 * If requested length is more that maximum possible length returns NULL
 *
 * Returns pointer to allocated buffer or NULL. len is set to allocated _usable_ buffer size
 *
 */
ldp_qmessage *
ldp_session_get_buffer(ldp_session *ls, int *len)
{
	int xlen;
	ldp_qmessage *lq;
	struct proto_ldp *pl;

	pl = ls->ln->pl;
	
	xlen = MAX(ls->max_pdu - LDP_PDU_HEADER_SIZE, 256);

	if (len != NULL)
	{
		if (*len > xlen)
			return NULL;
		*len = xlen;
	}

	if ((lq = mb_alloc(pl->pdu_pool, LDP_QHEADER_SIZE + xlen)) == NULL)
	{
		LOG_SS_ERR("Can't allocate buffer for sending message in session %d", ls->ln->lsr_id, ls->space_id);
		return NULL;
	}

	LOG_SS_DB2("Allocated packet size %d for sending ", xlen);

	lq->ls = ls;
	lq->messages = 1;

	return lq;
}

/**
 * ldp_session_sendto_all_butone - Broadcast message to all sessions except one
 * @pl: pointer to LDP protocol
 * @lq: pointer to message buffer
 * @one: pointer to LDP session to skip
 * @flags: sending option
 *
 * Message has to be prepared for sending (e.g. headers in network format, etc..)
 * BULK messages are not supported due to possibly different maximum PDU size.
 * Caller should free @lq after broadcasting if needed. Broadcasting is limited to sessions
 * with LDP_SSM_OPERATIONAL state.
 *
 * Returns: LDP_OK or LDP_ERR
 */
int
ldp_session_sendto_all_butone(struct proto_ldp *pl, ldp_qmessage *lq, ldp_session *one, int flags)
{
	ldp_session *ls, *ls_next;
	ldp_message_header *lmh;

	if (flags & LM_BULK)
		return LDP_ERR;

	lmh = (ldp_message_header *)LDP_QHEADER(lq);

	/* 
	 * Sending message _can_ (possibly) result in session withdraw, so we're using
	 * _DELSAFE macro here
	 */
	WALK_LIST_DELSAFE(ls, ls_next, pl->session_list)
	{
		ls = SKIP_BACK(ldp_session, n_list, ls);

		if ((ls == one) || (ls->state != LDP_SSM_OPERATIONAL))
			continue;

		/* Set per-session msgid */
		lmh->id = htons(ls->seq++);

		if (ldp_session_send_message(ls, lq, flags|LM_NOFREE) != LDP_OK)
			LOG_SS_VRB("Unable to broadcast message to session %R:%d due to ldp_send_message() error",
					ls->ln->lsr_id, ls->space_id);
	}

	if (!(flags & LM_NOFREE))
		mb_free(lq);

	return LDP_OK;
}

/**
 * ldp_session_send_message - Prepares message buffer for sending or send it
 * @ls: pointer to LDP session
 * @lq: pointer to message buffer
 * @flags: sending option
 *
 * If no @flags are not set, copies new qmessage and links it to sending queue updating
 * data counters. If immediate sending (LM_IMMEDIATE) is required and there are 
 * no other data queued we try to send message immediately without linking. 
 * If buffer contains multiple messages (LM_BULK) total messages length must be set
 * in lq->length field. @lq buffer is freed by default. This can be skipped by supplying
 * LM_NOFREE flag in @flags field.
 *
 * Message has to be prepared for sending (e.g. headers in network format, etc..).
 * At least LDP_SSM_INITIALIZED state is required for successful sending
 *
 */
int
ldp_session_send_message(ldp_session *ls, ldp_qmessage *lq, int flags)
{
	ldp_neighbour *ln = ls->ln;
	int length, msgtype;
	sock *sk;
	ldp_message_header *lmh;
	ldp_qmessage *tmp;
	struct proto_ldp *pl;
	uint8_t x1, x2;

	if (ls->state < LDP_SSM_INITIALIZED)
		return LDP_ERR;

	sk = ln->sk;
	pl = ln->pl;
	lmh = (ldp_message_header *)LDP_QHEADER(lq);

	if (flags & LM_BULK)
	{
		/*
		 * We can't easily check PDU limits here.
		 * Assume that LM_BULK users know what they do
		 */
		length = lq->length;
	}
	else
	{
		length = ntohs(lmh->length) + 4; /* sizeof(MessageId) */
		/* Check if we're within PDU sending limits */
		if ((length + LDP_PDU_HEADER_SIZE) > MAX(ls->max_pdu, 256))
		{
			LOG_SS_ERR("Message length %d is not within session PDU limits (%d)", length, MAX(ls->max_pdu, 256));
			if (!(flags & LM_NOFREE))
				mb_free(lq);
			return LDP_ERR;
		}

	}

	/* Get real message type from message header, see ldp_message_convert_header()  */
	x1 = *(((uint8_t *) lmh));
	x2 = *(((uint8_t *) lmh) + 1);
	msgtype = ((x1 & 0x7F) << 8) + x2;

	/* Check if we need/can send immediately */
	if ((flags & LM_IMMEDIATE) && ((ln->flags & LDP_FLAG_TRANSMIT_RUNNING) == 0))
	{
		/* Copy data */
		memcpy((byte *)sk->tbuf + LDP_PDU_HEADER_SIZE, lmh, length);
		ldp_session_send_buffer(ls, sk, length);
		ls->messages_out += lq->messages;
		LLOG_SS_DBG("Send %smessage type %s length %d", (flags & LM_BULK) ? "(BULK) " : "", 
				ldp_message_texttype(msgtype), length);
		if (!(flags & LM_NOFREE))
			mb_free(lq);
		return LDP_OK;
	}

	if (flags & LM_NOFREE)
	{
		/* Copy data */
		if ((tmp = mb_alloc(pl->pdu_pool, LDP_QHEADER_SIZE + length)) == NULL)
		{
			LLOG_SS_ERR("Can't allocate memory for queuing %s packet", ldp_message_texttype(msgtype));
			return LDP_ERR;
		}
		memcpy(tmp, lq, LDP_QHEADER_SIZE + length);
		lq = tmp;
	}

	/* Queue message for later processing */
	add_tail(&ln->packets_list, NODE lq);
	/* Update counters */
	ln->packets_length += length;
	lq->length = length;
	lq->ls = ls;
	LOG_SS_DBG("Queued %smessage type %s length %d total %d", (flags & LM_BULK) ? "(BULK) " : "",
			ldp_message_texttype(msgtype), length, ln->packets_length);

	/*
	 * Better way is to schedule transfer by timer
	 * not to do this in code, but, unfortunately,
	 * there are no hi-res timers available
	 */
	if ((ln->flags & LDP_FLAG_TRANSMIT_RUNNING) == 0)
	{
		/* Start timer.. ? */
	}


	return LDP_OK;
}

/*
 * ldp_session_shutdown_transport - Close LDP transport connection
 * @ln: pointer to neighbour
 *
 * Removes all logical sessions and TCP socket
 * Function is protected from recursive calling
 */
static void
ldp_session_shutdown_transport(ldp_neighbour *ln)
{
	ldp_session *ls, *ls_next;
	ldp_qmessage *lq;

	if (_IS_DELETED2(ln))
		return;
	_SET_DELETED2(ln);

	/* Shutdown transport session */
	LOG_SS_VRB("Shutting down transport LDP session %I <> %R:*", ln->local, ln->remote, ln->lsr_id);

	/* Free transport session socket queue */
	WALK_LIST_FIRST(lq, ln->packets_list)
	{
		rem_node(NODE lq);
		mb_free(lq);
	}

	/* Close all logical LDP sessions */
	WALK_LIST_DELSAFE(ls, ls_next, ln->session_list)
	{
		ls = SKIP_BACK(ldp_session, n_lsr, ls);
		LOG_SS_VRB("Preparing to delete session %R:%d from neighbour %I", ln->lsr_id, ls->space_id, ln->lsr_id);
		ldp_session_shutdown(ls);
	}

	/* Close and free TCP socket */
	if (ln->sk != NULL)
	{
		LOG_SS_DBG("Closing session socket %d (%p)", ln->sk->fd, ln->sk);
		rfree(ln->sk);
		ln->sk = NULL;
	}
	ln->tr_state = LDP_SESSION_TRANSPORT_NONE;
	/* Zero running transfer flags */
	ln->flags &= ~LDP_FLAG_TRANSMIT_RUNNING;
	/* 
	 * Remove active session flag since active/passive state can change in cases 
	 * where no transport TLV exists and router is connected to us with multiple
	 * LDP-enabled interfaces
	 */
	ln->flags &= ~LDP_FLAG_SESSION_ACTIVE;


	/* 
	 * Do not free transport data buffer due to:
	 * 1) for temporary problems we will allocate another buffer (possibly the same size later)
	 * 2) we don't know original buffer size (we do, but we have to have this definition in 2 different places)
	 */
	ln->dataoff = 0;

	_UNSET_DELETED2(ln);
}


/*
 * Setup listening LDP TCP socket.
 * Returns LDP_OK or LDP_ERR in case of error
 */
static int
ldp_session_setup_listen_socket()
{
	if ((tcp_sk = sk_new(&root_pool)) == NULL)
	{
		LOG_SS_ERR("Error creating LDP listening socket");
		return LDP_ERR;
	}

	tcp_sk->type = SK_TCP_PASSIVE;
	tcp_sk->sport = LDP_PORT;
	tcp_sk->flags = 0;
	
	tcp_sk->tos = IP_PREC_INTERNET_CONTROL;
	tcp_sk->rbsize = LDP_SOCK_RX_BUFFER;
	tcp_sk->tbsize = LDP_SOCK_TX_BUFFER;
	tcp_sk->rx_hook = ldp_tcp_incoming_hook;
	tcp_sk->err_hook = ldp_tcp_incoming_err_hook;

	if (sk_open(tcp_sk) != 0)
	{
		LLOG_SS_ERR("LDP listening socket setup failed: %m");
		goto err;
	}

	LLOG_SS_DBG("LDP listening socket created, fd=%d", tcp_sk->fd);

	return LDP_OK;
err:
	rfree(tcp_sk);
	return LDP_ERR;
}

/**
 * ldp_session_init_connect - Start transport session
 * @ls: pointer to session
 *
 * Sets up TCP socket to connect to LDP neighbour and
 * start TCP connect.
 *
 * Returns LDP_OK or LDP_ERR in case of error
 */
int
ldp_session_init_connect(ldp_session *ls)
{
	sock *sk = NULL;
	ldp_neighbour *ln = ls->ln;
	struct proto_ldp *pl = ln->pl;

	if (ln->sk != NULL)
	{
		LOG_SS_ERR("Session already set up for neighbour %R:*", ln->lsr_id);
		return LDP_ERR;
	}

	if ((sk = sk_new(MAIN_POOL(pl))) == NULL)
	{
		LOG_SS_ERR("Error creating LDP session socket");
		return LDP_ERR;
	}

	ln->sk = sk;
	/* Local and remote address selection is done in ldp_dispatch_hello() */
	ln->last_packet = CurrentTime;

	sk->type = SK_TCP_ACTIVE;
	sk->saddr = ln->local;
	sk->daddr = ln->remote;
	sk->dport = LDP_PORT;
	sk->rbsize = LDP_SOCK_RX_BUFFER;
	sk->tbsize = LDP_SOCK_TX_BUFFER;
	sk->tos = IP_PREC_INTERNET_CONTROL;
	sk->tx_hook = ldp_session_connect_hook;
	sk->err_hook = ldp_session_err_hook;
	sk->data = (void *)ln;

	if (sk_open(sk) != 0)
	{
		LOG_SS_ERR("Socket setup failed: %m");
		goto err;
	}

	LOG_SS_DBG("Initiating TCP connect from %I to %I to lsr %R:%d", ln->local, ln->remote, ln->lsr_id, ls->space_id);

	ls->state = LDP_SSM_NONEXISTENT;

	return LDP_OK;

err:
	rfree(sk);
	ln->sk = NULL;
	return LDP_ERR;
}

/**
 * ldp_session_prepare  - prepares LDP session for use
 * @ls: LDP session to prepare
 *
 *
 */
void
ldp_session_prepare(ldp_session *ls)
{
	timer *t;
	struct proto_ldp *pl = ls->ln->pl;
	struct ldp_config *c = PL_TO_CF(pl);

	/* Set default timer */
	ls->keep_interval = c->skeep_interval;

	t = tm_new(MAIN_POOL(pl));
	t->hook = ldp_session_hold_timeout_hook;
	t->data = ls;
	ls->hold_timer = t;

	t = tm_new(MAIN_POOL(pl));
	t->hook = ldp_session_send_keepalive_hook;
	t->data = ls;
	ls->send_timer = t;

	/* Init all lists */
	init_list(&ls->addr_list);
	init_list(&ls->prefix_list);
	init_list(&ls->label_list);
}

/*
 * ldp_session_send_buffer - crafts PDU and send requested messages
 * @ls: pointer to LDP session
 * @sk: socket to send to
 * @length: total messages length
 *
 * Function fills in PDU header in the beginning of socket buffer and
 * tries to send this data
 *
 */
static int
ldp_session_send_buffer(ldp_session *ls, sock *sk, int length)
{
	int res;

	/* Fill header and get full PDU length */
	length = ldp_pdu_put_header(sk->tbuf, length, ls->space_id);

	if ((res = sk_send(sk, length)) == 1)
	{
		/* Packet sent, let's update counters */
		ls->last_sent = CurrentTime;
	
		/* Restart our keepalive timer */
		tm_stop(ls->send_timer);
		tm_start(ls->send_timer, ls->keep_interval / 3);
	}
	else
	{
		/* 
		 * If sk_send returns 0 we will get status update via
		 * tx_ or err_hook. No more packets should be send for now
		 */
		ls->ln->flags |= LDP_FLAG_TRANSMIT_RUNNING;
	}

	return res;
}

#define LDP_SOCK_PROCESSING_LIMIT	400
/*
 * ldp_session_fire_transfer - send packet queue to given neighbour
 * @ln: pointer to LDP nighbour
 *
 */
static void
ldp_session_fire_transfer(ldp_neighbour *ln)
{
	int length, i = 0;
	ldp_qmessage *lq;
	ldp_session *ls;
	sock *sk;

	if (EMPTY_LIST(ln->packets_list))
		return;

	if (ln->flags & LDP_FLAG_TRANSMIT_RUNNING)
		return;

	LOG_SS_DBG("TRANSMITTER started for neighbour *, queue length: %d", ln->lsr_id, ln->packets_length);

	lq = (ldp_qmessage *)HEAD(ln->packets_list);
	/* Set initial session id to first element value */
	ls = lq->ls;

	sk = ln->sk;
	length = LDP_PDU_HEADER_SIZE;

	WALK_LIST_FIRST(lq, ln->packets_list)
	{
		/* PDU can't contain messages from different labelspaces */
		if ((lq->ls != ls) || (length + lq->length >= ls->max_pdu))
		{
			LOG_SS_DBG("Send session PDU length %d", length);
			/* We have to send current packet */
			if (ldp_session_send_buffer(ls, sk, length) == 0)
			{
				/* Packet is queued for transmit, let's stop now */
				return;
			}

			if (i++ >= LDP_SOCK_PROCESSING_LIMIT)
				return;

			ls = lq->ls;
			length = LDP_PDU_HEADER_SIZE;
		}

		memcpy(sk->tbuf + length, LDP_QHEADER(lq), lq->length);
		length += lq->length;
		ln->packets_length -= lq->length;
		ls->messages_out += lq->messages;
		rem_node(NODE lq);
		mb_free(lq);
	}

	/* Send remaining buffer */
	if (length > 0)
	{
		LOG_SS_DBG("Send session PDU length %d", length);
		ldp_session_send_buffer(ls, sk, length);
	}
}


/**
 * ldp_session_get_role - Determines active or passive role in transport connection establishment
 * @li: pointer to interface PDU is received on
 * @lm: pointer to received HELLO
 *
 * Returns 1 if mode is active (have to initiate TCP connect) or 0 for PASSIVE
 * See RFC5038 Clause 2.5.2 for more information
 */
static enum LDP_SESS_ROLE
ldp_session_get_role(struct ldp_iface *li, ip_addr remote)
{
	struct proto_ldp *pl = li->pl;
	ip_addr local;

	local = pl->tcp_saddr;

	LOG_SS_DBG("Comparing session adresses: %I <> %I", local, remote);

	if (ipa_compare(local, remote) >= 0)
	{
		LOG_SS_DBG("Session is ACTIVE");
		return LDP_SESSION_ACTIVE;
	}

	LOG_SS_DBG("Session is PASSIVE");
	return LDP_SESSION_PASSIVE;
}

static int
ldp_session_preprocess(ldp_neighbour *ln, byte *data, int size)
{
	struct proto_ldp *pl = ln->pl;
	ldp_pdu_header lph;
	ldp_session *ls;
	int length, res;

	LLOG_SS_DBG("Called with datasize %d for neighbour %R:*", size, ln->lsr_id);

	res = ldp_pdu_preprocess(pl, ln->remote, ln->datasize, data, size, ln->data, ln->datasize, ln->dataoff, &length);
	ln->dataoff += length;

	/* Parse error */
	if (res == -1)
	{
		/* 
		 * Try to determine session to close. received data length is not enough to determine 
		 * logical session, close transport session 
		 */

		if (ln->dataoff < LDP_PDU_HEADER_SIZE)
		{
			LOG_SS_VRB("Bad data received from %I, closing transport session", ln->remote);
			ldp_session_shutdown_transport(ln);
			return -1;
		}

		/* Header data read, try to determine session */
		ldp_pdu_get_header(ln->data, &lph);

		if ((ls = ldp_session_find(pl, ln->lsr_id, lph.space_id)) == NULL)
		{
			LOG_SS_VRB("Bad data with unknown labelspace %d received from %R , closing transport session", lph.space_id, ln->lsr_id);
			ldp_session_shutdown_transport(ln);
			return -1;
		}

		/* Close logical session */
		LOG_SS_VRB("Bad PDU data received from %I, closing session with %R:%d", ln->remote, ln->lsr_id, ls->space_id);
		ldp_session_shutdown(ls);
		return -1;
	}

	/* Some data arrived, PDU is still not complete */
	if (res == 0)
		return length;

	/* Convert PDU header to host format */
	ldp_pdu_get_header(ln->data, &lph);

	/* Associate with logical session */
	if ((ls = ldp_session_find(pl, ln->lsr_id, lph.space_id)) == NULL)
	{
		LOG_SS_DBG("PDU with unknown labelspace %d received from %R:*, ignoring", lph.space_id, ln->lsr_id);
		return length;
	}

	/* Check if received PDU is within session consiled PDU limits */
	if (ls->max_pdu < length)
	{
		LOG_SS_VRB("Too big PDU (length %d > max %d) received in session %R:%d, ignoring", length, ls->max_pdu, ln->lsr_id, ls->space_id);
		return length;
	}

	ldp_pdu_dispatch(ln, ln->data, ln->dataoff, LDP_PDU_SESSION, ls);

	ln->dataoff = 0;

	return length;
}

/* VARIOUS HOOKS */

/**
 * ldp_session_connect_hook - handles outgoing connection established event
 * @sk: TCP socket
 */
static void
ldp_session_connect_hook(sock *sk)
{
	ldp_neighbour *ln = (ldp_neighbour *)sk->data;
	struct proto_ldp *pl = ln->pl;
	ldp_session *ls, *ls_next;

	DEBUG_PROTO(pl);

	ln->tr_state = LDP_SESSION_TRANSPORT_ACTIVE;
	/* Setup RX/TX hooks */
	sk->rx_hook = ldp_session_rx_hook;
	sk->tx_hook = ldp_session_tx_hook;

	LOG_SS_DBG("transport connection with %R:* done, sending INIT PDU for every session", ln->lsr_id);
	WALK_LIST_DELSAFE(ls, ls_next, ln->session_list)
	{
		ls = SKIP_BACK(ldp_session, n_lsr, ls);
		/* Let's transmit INIT and set new state*/
		ldp_session_set_state(ls, LDP_SSM_INITIALIZED);
		ldp_session_send_init(ls);
		ldp_session_set_state(ls, LDP_SSM_OPENSENT);
	}

	UNDEBUG_PROTO();
}

/**
 * ldp_session_incoming_hook - handle an incoming connection
 * @sk: TCP socket
 * @dummy: unused
 *
 * This function serves as a socket hook for accepting of new LDP
 * connections. 
 */
static int
ldp_tcp_incoming_hook(sock *sk, int dummy UNUSED)
{
	struct proto_ldp *pl, *pl_next;
	ldp_neighbour *ln, *ln_next;
	ldp_session *ls, *ls_next;
	int i = 0;

	DBG("Incoming LDP connection from %I port %d", sk->daddr, sk->dport);

	/* Loop thru all sessions and set OPENREC flag to all appropriate since all such sessions are PASSIVE */
	WALK_LIST_DELSAFE(pl, pl_next, pl_list)
	{
		pl = SKIP_BACK(struct proto_ldp, pl_list, pl);
		DEBUG_PROTO(pl);
		WALK_LIST_DELSAFE(ln, ln_next, pl->neighbour_list)
		{
			ln = SKIP_BACK(ldp_neighbour, n_list, ln);
	
			if (!ipa_equal(ln->remote, sk->daddr))
				continue;

			/* Check if neighbour doesn't have transport session */
			if (ln->tr_state != LDP_SESSION_TRANSPORT_NONE)
			{
				LOG_SS_DBG("Found neighbour with active transport, skipping");
				continue;
			}

			LOG_SS_DBG("Found neigbour %R:*", ln->lsr_id);
			//LOG_SS_DBG("rxbuf=%d txbuf=%d", sk->rbsize, sk->tbsize);
			i++;
			break;
		}
		UNDEBUG_PROTO();
	}

	if (i == 0)
	{
		DBG("No session associated with %I, closing", sk->daddr);
		rfree(sk);
		return 0;
	}

	/* Neighbour found, so pl and ln should hold valid values */
	DEBUG_PROTO(pl);

	WALK_LIST_DELSAFE(ls, ls_next, ln->session_list)
	{
		ls = SKIP_BACK(ldp_session, n_lsr, ls);
		ldp_session_set_state(ls, LDP_SSM_OPENREC);
	}

	/* Set up hooks and data */
	ln->sk = sk;
	sk->data = ln;
	sk->rx_hook = ldp_session_rx_hook;
	sk->tx_hook = ldp_session_tx_hook;
	sk->err_hook = ldp_session_err_hook;

	UNDEBUG_PROTO();

	return 1;
}

static void
ldp_tcp_incoming_err_hook(sock *sk, int err)
{
	struct ldp_iface *li = (struct ldp_iface *) (sk->data);
	struct proto_ldp *pl;

	if (!li)
		return;

	pl = li->pl;

	DEBUG_PROTO(pl);
	
	if (err == ECONNABORTED)
	{
		/* New connection aborted */
		LOG_VRB("New incoming TCP connection aborted");
	}
	else
	{
		/* 
		 * Problems with listening socket.
		 * XXX: We should notify user somehow
		 */
		LOG_SS_ERR("Listening socket down: %M", err);
	}

	UNDEBUG_PROTO();
}

/**
 * ldp_session_rx_hook
 * @sk: socket we received the packet.
 * @size: size of the packet
 *
 * This is the entry point for all session messages from neighbors.
 */
static int
ldp_session_rx_hook(sock *sk, int size)
{
	ldp_neighbour *ln = (ldp_neighbour *) sk->data;
	struct proto_ldp *pl = ln->pl;
	int length, off = 0, error = 1;

	DEBUG_PROTO(pl);

	LOG_SS_DBG("RX hook (src %I, dst %I, size %d)", sk->daddr, sk->laddr, size);

	/* Some simple parsing logic has to be kept here */
	while (42)
	{
		if ((length = ldp_session_preprocess(ln, sk->rbuf + off, size)) == -1)
		{
			LOG_SS_VRB("ldp_preprocess_session() returned error for data from neighbour %R:*", ln->lsr_id);
			break;
		}

		/* 
		 * It is possible that we receive valid LDP NOTIFICATION message with fatal bit set.
		 * In most common case this will cause us to close transport session (e.g. close) socket
		 * If this was not the last message from PDU we will get core on next cycle
		 */
		if (ln->sk == NULL)
		{
			LOG_SS_DBG("Transport session with %R:* closed, skipping all PDUs..", ln->lsr_id);
			break;
		}

		if (length >= size)
		{
			if (off > 0)
			{
				/* 
				 * We've distpatched one or more PDUs and now we're 
				 * at the beginning of next PDU but it has't received yet.
				 * We're altering sk buffer/buffer length to make
				 * buffer begin with this PDU begingging..
				 * We're also returning 0 to indicate that data should be kept
				 *
				 * XXX: Some internal knowledge about sk is used 
				 * here. 
				 * */
				memmove(sk->rbuf, sk->rbuf + off, size);
				sk->rpos -= off;
			}
			/* unix/io.c:1406. Return 0 not to set rpos back to rbuf (e.g zero all data) */
			error = 0;
			break;
		}

		size -= length;
		off += length;
	}

	UNDEBUG_PROTO();

	return error;
}


static void
ldp_session_check_tranfers(timer *t)
{
	struct proto_ldp *pl = t->data;
	ldp_neighbour *ln, *ln_next;

	DEBUG_PROTO(pl);
	UNDEBUG_PROTO();
	return;

	WALK_LIST_DELSAFE(ln, ln_next, pl->neighbour_list)
	{
		ln = SKIP_BACK(ldp_neighbour, n_list, ln);

/*
		LDBG("%p transmitter check running.. neighbour %I length=%d transmit %s", 
				pl, ln->lsr_id, ln->packets_length, (ln->flags & LDP_FLAG_TRANSMIT_RUNNING) ? "RUNNING" : "STOPPED");
*/
		if (ln->tr_state != LDP_SESSION_TRANSPORT_ACTIVE)
			continue;

		if (ln->packets_length == 0)
			continue;

		if (ln->flags & LDP_FLAG_TRANSMIT_RUNNING)
			continue;

		ldp_session_fire_transfer(ln);
	}

	UNDEBUG_PROTO();
}

static void
ldp_session_tx_hook(sock *sk)
{
	ldp_neighbour *ln = sk->data;
	struct proto_ldp *pl = ln->pl;

	DEBUG_PROTO(pl);
	
	LOG_SS_DB2("TX hook returned, running transmitter for neighbour *", ln->lsr_id);
	ln->flags &= ~LDP_FLAG_TRANSMIT_RUNNING;

	/* All data transmitted successfully. Let's check if we need to send more */
	ldp_session_fire_transfer(ln);

	UNDEBUG_PROTO();
}

static void
ldp_session_send_keepalive_hook(timer *t)
{
	ldp_session *ls = t->data;
	DEBUG_PROTO(ls->ln->pl);

	LOG_SS_DBG("Sending keepalive for LDP session %R:%d", ls->ln->lsr_id, ls->space_id);

	ldp_session_send_keepalive(ls);
	UNDEBUG_PROTO();
}

static void
ldp_session_hold_timeout_hook(timer *t)
{
	ldp_session *ls = t->data;
	DEBUG_PROTO(ls->ln->pl);

	LOG_SS_ERR("LDP session %R:%d is down (Keepalive not received)", ls->ln->lsr_id, ls->space_id);

	ldp_session_shutdown(ls);
	UNDEBUG_PROTO();
}

static void
ldp_session_err_hook(sock *sk, int err)
{
	ldp_neighbour *ln = (ldp_neighbour *) (sk->data);
	struct proto_ldp *pl;

	if (!ln)
		return;

	pl = ln->pl;

	DEBUG_PROTO(pl);

	LOG_SS_VRB("Socket error: %M in transport session with %R:*", err, ln->lsr_id);

	if ((sk != NULL) && (ln->sk != NULL) && (sk != ln->sk))
	{
		LOG_SS_ERR("Sockets are not equal in session with *: sk_fd=%d ln->sk_fb=%d", ln->lsr_id, sk->fd, ln->sk->fd);
		rfree(sk);
	}

	/* Remove pointer to neighbour */
	sk->data = NULL;

	/* We need to mark transport session inactive to skip sending notificastions for every session */
	ln->tr_state = LDP_SESSION_TRANSPORT_NONE;
	ldp_session_shutdown_transport(ln);

	UNDEBUG_PROTO();
}

