/*
 *	BIRD -- Table-to-Table Routing Protocol a.k.a Pipe
 *
 *	(c) 1999--2000 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

/**
 * DOC: Pipe
 *
 * The Pipe protocol is very simple. It just connects to two routing tables
 * using proto_add_announce_hook() and whenever it receives a rt_notify()
 * about a change in one of the tables, it converts it to a rte_update()
 * in the other one.
 *
 * To avoid pipe loops, Pipe keeps a `being updated' flag in each routing
 * table.
 */

#undef LOCAL_DEBUG

#include "nest/bird.h"
#include "nest/iface.h"
#include "nest/protocol.h"
#include "nest/route.h"
#include "nest/cli.h"
#include "conf/conf.h"
#include "filter/filter.h"
#include "lib/string.h"

#include "pipe.h"

static void
pipe_rt_notify(struct proto *P, rtable *src_table, net *n, rte *new, rte *old, ea_list *attrs)
{
  struct pipe_proto *p = (struct pipe_proto *) P;
  rtable *dest;
  struct proto *src;
  struct announce_hook *h;
  struct proto_stats *stats, *old_stats;

  net *nn;
  rte *e;
  rta a;

  if (!new && !old)
    return;

  if (src_table == P->table)
    {
      dest = p->peer;
      h = p->a;
      stats = &P->stats;
    }
  else
    {
      dest = P->table;
      h = P->thook;
      stats = &p->peer_stats;
    }

  if (dest->pipe_busy)
    {
      log(L_ERR "Pipe loop detected when sending %I/%d to table %s",
	  n->n.prefix, n->n.pxlen, dest->name);
      return;
    }
  nn = net_get(dest, n->n.prefix, n->n.pxlen);
  if (new)
    {
      memcpy(&a, new->attrs, sizeof(rta));

      if (p->mode == PIPE_OPAQUE)
	{
	  a.proto = &p->p;
	  a.source = RTS_PIPE;
	}

      a.aflags = 0;
      a.eattrs = attrs;
      a.hostentry = NULL;
      e = rte_get_temp(&a);
      e->net = nn;
      e->pflags = 0;

      if (p->mode == PIPE_TRANSPARENT)
	{
	  /* Copy protocol specific embedded attributes. */
	  memcpy(&(e->u), &(new->u), sizeof(e->u));
	  e->pref = new->pref;
	  e->pflags = new->pflags;
	}

      src = new->attrs->proto;
    }
  else
    {
      e = NULL;
      src = old->attrs->proto;
    }

  src_table->pipe_busy = 1;
  /*
   * FIXME: Trick with stats pointer is needed to produce
   * valid numbers on protocol shutdown.
   */
  old_stats = h->stats;
  h->stats = stats;
  do_rte_update(dest, h, nn, &p->p, (p->mode == PIPE_OPAQUE) ? &p->p : src, e);
  h->stats = old_stats;
  src_table->pipe_busy = 0;
}

static int
pipe_import_control(struct proto *P, rte **ee, ea_list **ea UNUSED, struct linpool *p UNUSED)
{
  struct proto *pp = (*ee)->sender;

  if (pp == P)
    return -1;	/* Avoid local loops automatically */
  return 0;
}

static int
pipe_reload_routes(struct proto *P)
{
  /*
   * Because the pipe protocol feeds routes from both routing tables
   * together, both directions are reloaded during refeed and 'reload
   * out' command works like 'reload' command. For symmetry, we also
   * request refeed when 'reload in' command is used.
   * We have to remove import limit here, too
   */
  proto_clear_limits(P, PFLAG_ILIMIT|PFLAG_ILIMIT_BLOCK|PFLAG_ELIMIT|PFLAG_ELIMIT_BLOCK);
  proto_request_feeding(P);
  return 1;
}

static int
pipe_start(struct proto *P)
{
  struct pipe_proto *p = (struct pipe_proto *) P;
  struct announce_hook *a;

  /* Clean up the secondary stats */
  bzero(&p->peer_stats, sizeof(struct proto_stats));

  /* Lock the peer table, unlock is handled in pipe_cleanup() */
  rt_lock_table(p->peer);

  /* Connect the protocol also to the peer routing table. */
  a = proto_add_announce_hook(P, p->peer, P->accept_ra_types);
  p->a = a;

  /* Set up filters and stats */
  a->stats = &p->peer_stats;

  return PS_UP;
}

static void
pipe_cleanup(struct proto *P)
{
  struct pipe_proto *p = (struct pipe_proto *) P;
  rt_unlock_table(p->peer);
}

static struct proto *
pipe_init(struct proto_config *C)
{
  struct pipe_config *c = (struct pipe_config *) C;
  struct proto *P = proto_new(C, sizeof(struct pipe_proto));
  struct pipe_proto *p = (struct pipe_proto *) P;

  p->peer = c->peer->table;
  p->mode = c->mode;
  P->accept_ra_types = (p->mode == PIPE_OPAQUE) ? RA_OPTIMAL : RA_ANY;
  P->rt_notify = pipe_rt_notify;
  P->rt_notify_any = pipe_rt_notify;
  P->import_control = pipe_import_control;
  P->reload_routes = pipe_reload_routes;

  return P;
}

static void
pipe_postconfig(struct proto_config *C)
{
  struct pipe_config *c = (struct pipe_config *) C;

  if (!c->peer)
    cf_error("Name of peer routing table not specified");
  if (c->peer == C->table)
    cf_error("Primary table and peer table must be different");
}

static int
pipe_reconfigure(struct proto *P, struct proto_config *new)
{
  struct announce_hook *a, *pa;
  struct pipe_proto *p = (struct pipe_proto *)P;
  struct pipe_config *o = (struct pipe_config *) P->cf;
  struct pipe_config *n = (struct pipe_config *) new;

  if ((o->peer->table != n->peer->table) || (o->mode != n->mode))
    return 0;

  /*
   * reconfigure can be called in any protocol status
   */
  if (!((a = P->thook) && (pa = p->a)))
    return 1;

  /* Update filters */
  /*
   * Export filter remains as is and import filter moves to
   * peer table
   *
   * /------\                                      /------\
   * |      | --- (OUT)(1)            (OUT)(3) --- |      |
   * | MAIN |               | PIPE |               | PEER |
   * |      | ---- (IN)(2)             (IN)(4) --- |      |
   * \------/                                      \------/
   *
   * When new route is announced on MAIN table it gets checked by
   * export filter (1), and, after that, it is announced to peer table
   * via rte_update. import filter (4) is called. When new route is
   * annouced in PEER table (3) and (2) are used. Oviously, there is
   * no need in filtering the same route twice, so (4) and (2) filters
   * should be set to pass all routes.
   *
   * User can configure (1) and (2) filters so we move filter (2) to
   * (3) and set (2) and (4) to FILTER_ACCEPT
   *
   * This is the right place to do it since
   * 1) configure hook is called exactly before initial feeding
   * 2) it is (now) called exactly after setting new filters in proto_reconfigure
   */

  /*
   * Check for changed filters:
   * Import and export filters are checked in proto_reconfigure by
   * comparing old/new configurations so no additional checks are
   * required here
   */

  pa->out_filter = a->in_filter;
  a->in_filter = FILTER_ACCEPT;
  pa->in_filter = FILTER_ACCEPT;

  /*
   * Update limits
   * Note that import/export limit can have different functionality
   * (for example, RESTART action can't be used in export filter)
   * so this trick may impose some usage limitation.
   */
  pa->out_limit = a->in_limit;
  a->in_limit = NULL;

  return 1;
}

void
pipe_show_stats(struct proto *P)
{
  struct pipe_proto *p = (struct pipe_proto *) P;
  struct proto_stats *s1 = &P->stats;
  struct proto_stats *s2 = &p->peer_stats;

  /*
   * Pipe stats (as anything related to pipes) are a bit tricky. There
   * are two sets of stats - s1 for routes going from the primary
   * routing table to the secondary routing table ('exported' from the
   * user point of view) and s2 for routes going in the other
   * direction ('imported' from the user point of view).
   *
   * Each route going through a pipe is, technically, first exported
   * to the pipe and then imported from that pipe and such operations
   * are counted in one set of stats according to the direction of the
   * route propagation. Filtering is done just in the first part
   * (export). Therefore, we compose stats for one directon for one
   * user direction from both import and export stats, skipping
   * immediate and irrelevant steps (exp_updates_accepted,
   * imp_updates_received, imp_updates_filtered, ...)
   */

  cli_msg(-1006, "  Routes:         %u imported, %u exported", 
	  s2->imp_routes, s1->imp_routes);
  cli_msg(-1006, "  Route change stats:     received   rejected   filtered    ignored   accepted");
  cli_msg(-1006, "    Import updates:     %10u %10u %10u %10u %10u",
	  s2->exp_updates_received, s2->exp_updates_rejected + s2->imp_updates_invalid,
	  s2->exp_updates_filtered, s2->imp_updates_ignored, s2->imp_updates_accepted);
  cli_msg(-1006, "    Import withdraws:   %10u %10u        --- %10u %10u",
	  s2->exp_withdraws_received, s2->imp_withdraws_invalid,
	  s2->imp_withdraws_ignored, s2->imp_withdraws_accepted);
  cli_msg(-1006, "    Export updates:     %10u %10u %10u %10u %10u",
	  s1->exp_updates_received, s1->exp_updates_rejected + s1->imp_updates_invalid,
	  s1->exp_updates_filtered, s1->imp_updates_ignored, s1->imp_updates_accepted);
  cli_msg(-1006, "    Export withdraws:   %10u %10u        --- %10u %10u",
	  s1->exp_withdraws_received, s1->imp_withdraws_invalid,
	  s1->imp_withdraws_ignored, s1->imp_withdraws_accepted);
}

static void
pipe_copy_config(struct proto_config *dest, struct proto_config *src)
{
  /* Just a shallow copy, not many items here */
  proto_copy_rest(dest, src, sizeof(struct pipe_config));
}

static void
pipe_get_status(struct proto *P, byte *buf)
{
  struct pipe_proto *p = (struct pipe_proto *) P;

  bsprintf(buf, "%c> %s", (p->mode == PIPE_OPAQUE) ? '-' : '=', p->peer->name);
}



struct protocol proto_pipe = {
  name:		"Pipe",
  template:	"pipe%d",
  preference:	DEF_PREF_PIPE,
  postconfig:	pipe_postconfig,
  init:		pipe_init,
  start:	pipe_start,
  cleanup:	pipe_cleanup,
  reconfigure:	pipe_reconfigure,
  copy_config:  pipe_copy_config,
  get_status:	pipe_get_status,
};
